import asyncio
import json
from json import JSONDecodeError

from aiohttp import web

from guapow.common.dto import OptimizationRequest


async def optimize(request: web.Request):
    body = request.get('decrypted_text')

    if body is None:
        body = (await request.text()).strip()

    try:
        opt_req = OptimizationRequest(**json.loads(body))
    except (JSONDecodeError, TypeError):
        request.app.logger.warning('Invalid request')
        return web.Response(status=400)

    if opt_req.user_name:
        opt_req.user_id = request.app['allowed_users'].get(opt_req.user_name)

        if opt_req.user_id is None:
            request.app.logger.info(f"Request not allowed for user '{opt_req.user_name}' (pid={opt_req.pid})")
            return web.Response(status=401)

    if opt_req.is_valid():

        request.app.logger.info(f'New request: {opt_req}')

        if await request.app['queue'].add_pid(opt_req.pid):
            asyncio.create_task(request.app['handler'].handle(opt_req))
            return web.Response(status=200)
        else:
            request.app.logger.info(f"Repeated request for process {opt_req.pid}. Ignoring it.")
            return web.Response(status=202)
    else:
        request.app.logger.warning(f'Invalid request: {opt_req}')
        return web.Response(status=400)



import re
import shutil
from abc import ABC, abstractmethod
from asyncio import Lock
from logging import Logger
from shutil import which
from typing import Optional, Dict, Set, Tuple

from guapow.common import system
from guapow.common.system import run_async_user_process

RE_COMPOSITOR_NAME = re.compile(r'compositor\s*:\s*(.+)\s')


class WindowCompositor(ABC):
    """ Compositing Window Manager """

    def __init__(self, logger: Logger):
        self._log = logger
        self._lock = Lock()

    def lock(self) -> Lock:
        return self._lock

    @abstractmethod
    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        pass

    @abstractmethod
    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        pass

    @abstractmethod
    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        pass

    @abstractmethod
    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        """
        return: tuple with the first element as the status (enabled/disabled) and the second element a context object that can
        be passed to the other calls
        """
        pass

    @abstractmethod
    def get_name(self) -> str:
        pass


class WindowCompositorWithCLI(WindowCompositor):
    """
    Window compositor that has a command line interface to interact with.
    """
    def __init__(self, name: str, enable_cmd: str, disable_cmd: str, is_enable_cmd, logger: Logger):
        super(WindowCompositorWithCLI, self).__init__(logger)
        self._name = name
        self._enable_cmd = enable_cmd
        self._disable_cmd = disable_cmd
        self._is_enable_cmd = is_enable_cmd
        self._commands = {c.split(' ')[0].strip() for c in {enable_cmd, disable_cmd, is_enable_cmd}}

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        if self._commands:
            for c in self._commands:
                if not shutil.which(c):
                    return False, f"'{c}' is not installed"

        return True, None

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        code, output = await system.run_async_user_process(self._enable_cmd, user_id, user_env)

        if code == 0:
            return True
        else:
            error_output = output.replace('\n', ' ') if output else ''
            self._log.error(f"Could not enable {self.get_name()}. Command ({self._enable_cmd}) failed. Output: {error_output}")
            return False

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        code, output = await system.run_async_user_process(self._disable_cmd, user_id, user_env)

        if code == 0:
            return True
        else:
            error_output = output.replace('\n', ' ') if output else ''
            self._log.error(f"Could not disable {self.get_name()}. Command ({self._disable_cmd}) failed. Output: {error_output}")
            return False

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        exitcode, output = await system.run_async_user_process(self._is_enable_cmd, user_id, user_env)

        if exitcode == 0:
            if not output:
                self._log.error(f"Could not determine if {self.get_name()} is enabled. No output from command: {self._is_enable_cmd}")
                return

            state_str = output.strip().lower()
            if state_str == 'true':
                return True
            elif state_str == 'false':
                return False
            else:
                log_output = output.replace('\n', ' ')
                self._log.warning(f'Could not determine if {self.get_name()} is enabled. Unknown output from command "{self._is_enable_cmd}": {log_output}')
        else:
            log_output = output.replace('\n', ' ') if output is not None else ' '
            self._log.error(f'Could not determine if {self.get_name()} is enabled. Command "{self._is_enable_cmd}" failed (exitcode={exitcode}). Output: {log_output}')

    def get_name(self) -> str:
        return self._name


class KWinCompositor(WindowCompositor):

    def __init__(self, logger: Logger):
        super(KWinCompositor, self).__init__(logger)
        self._compositor = WindowCompositorWithCLI(name='KWin',
                                                   enable_cmd='qdbus org.kde.KWin /Compositor resume',
                                                   disable_cmd='qdbus org.kde.KWin /Compositor suspend',
                                                   is_enable_cmd='qdbus org.kde.KWin /Compositor org.kde.kwin.Compositing.active',
                                                   logger=logger)

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        return self._compositor.can_be_managed()

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.enable(user_id=user_id, user_env=user_env, context=context)

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.disable(user_id=user_id, user_env=user_env, context=context)

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        return await self._compositor.is_enabled(user_id=user_id, user_env=user_env, context=context)

    def get_name(self) -> str:
        return self._compositor.get_name()


class Xfwm4Compositor(WindowCompositor):

    def __init__(self, logger: Logger):
        super(Xfwm4Compositor, self).__init__(logger)
        self._compositor = WindowCompositorWithCLI(name='Xfwm4',
                                                   enable_cmd='xfconf-query --channel=xfwm4 --property=/general/use_compositing --set=true',
                                                   disable_cmd='xfconf-query --channel=xfwm4 --property=/general/use_compositing --set=false',
                                                   is_enable_cmd='xfconf-query --channel=xfwm4 --property=/general/use_compositing',
                                                   logger=logger)

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        return self._compositor.can_be_managed()

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.enable(user_id=user_id, user_env=user_env, context=context)

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.disable(user_id=user_id, user_env=user_env, context=context)

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        return await self._compositor.is_enabled(user_id=user_id, user_env=user_env, context=context)

    def get_name(self) -> str:
        return self._compositor.get_name()


class MarcoCompositor(WindowCompositor):

    def __init__(self, logger: Logger):
        super(MarcoCompositor, self).__init__(logger)
        self._compositor = WindowCompositorWithCLI(name='Marco',
                                                   enable_cmd='gsettings set org.mate.Marco.general compositing-manager true',
                                                   disable_cmd='gsettings set org.mate.Marco.general compositing-manager false',
                                                   is_enable_cmd='gsettings get org.mate.Marco.general compositing-manager',
                                                   logger=logger)

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        return self._compositor.can_be_managed()

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.enable(user_id=user_id, user_env=user_env, context=context)

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.disable(user_id=user_id, user_env=user_env, context=context)

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        return await self._compositor.is_enabled(user_id=user_id, user_env=user_env, context=context)

    def get_name(self) -> str:
        return self._compositor.get_name()


class WindowCompositorNoCLI(WindowCompositor):

    def __init__(self, name: str, process_name: str, logger: Logger):
        super(WindowCompositorNoCLI, self).__init__(logger)
        self._name = name
        self._process_name = process_name.strip()
        self._re_process_name: Optional[re.Pattern] = None

    def get_name(self) -> str:
        return self._name

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        if not shutil.which(self._process_name):
            return False, f"'{self._process_name}' is not installed"

        return True, None

    async def enable(self, user_id: Optional[int], user_env: Optional[Dict[str, str]], context: dict) -> bool:
        enable_cmd = context.get('cmd')

        if not enable_cmd:
            self._log.error(f"Enable command not available on context for compositor '{self.get_name()}'")
            return False

        code, output = await system.run_async_user_process(cmd=enable_cmd, user_id=user_id, user_env=user_env)

        if code == 0:
            return True

        error_log = output.replace('\n', ' ') if output else ''
        self._log.error(f"Could not start window compositor '{self.get_name()}'. Command '{enable_cmd}' failed. Output: {error_log}")
        return False

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        pid = context.get('pid')

        if pid is None:
            self._log.error(f"Window compositor {self.get_name()} process id could not be found on the context ({context}). It will not be disabled.")
            return False

        code, output = await system.async_syscall('kill -9 {}'.format(pid))

        if code == 0:
            return True

        error_log = output.replace('\n', ' ') if output else ''
        self._log.error(f"Could stop window compositor process '{self._process_name}' (pid={pid}). Output: {error_log}")
        return False

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        if self._re_process_name is None:
            self._re_process_name = re.compile(r'^{}$'.format(self._process_name))

        proc_data = await system.find_process_by_name(self._re_process_name)

        if proc_data:
            context['pid'] = proc_data[0]
            context['cmd'] = proc_data[1]
            return True

        return False


class PicomCompositor(WindowCompositor):

    def __init__(self, name: str, logger: Logger):
        super(PicomCompositor, self).__init__(logger)
        self._compositor = WindowCompositorNoCLI(name=name.capitalize(), process_name=name, logger=logger)

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        return self._compositor.can_be_managed()

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.enable(user_id, user_env, context)

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.disable(user_id, user_env, context)

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        return await self._compositor.is_enabled(user_id, user_env, context)

    def get_name(self) -> str:
        return self._compositor.get_name()


class CompizCompositor(WindowCompositor):

    def __init__(self, logger: Logger):
        super(CompizCompositor, self).__init__(logger)
        self._compositor = WindowCompositorNoCLI(name='Compiz', process_name='compiz', logger=logger)

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        return self._compositor.can_be_managed()

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.enable(user_id, user_env, context)

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._compositor.disable(user_id, user_env, context)

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        return await self._compositor.is_enabled(user_id, user_env, context)

    def get_name(self) -> str:
        return self._compositor.get_name()


class NvidiaCompositor(WindowCompositor):

    def __init__(self, logger: Logger):
        super(NvidiaCompositor, self).__init__(logger)
        self._re_attrs: Optional[re.Pattern] = None
        self._main_cmd = 'nvidia-settings'

    def can_be_managed(self) -> Tuple[bool, Optional[str]]:
        if not shutil.which(self._main_cmd):
            return False, f"'{self._main_cmd}' is not installed"

        return True, None

    def _get_re_attrs(self) -> re.Pattern:
        if self._re_attrs is None:
            self._re_attrs = re.compile(r'((Force(Full)?CompositionPipeline)\s*=\s*\w+)', re.IGNORECASE)

        return self._re_attrs

    def extract_attributes(self, string: str) -> Optional[Set[str]]:
        matches = self._get_re_attrs().findall(string)

        if matches:
            return {*{m[1].strip().lower(): m[1].strip() for m in matches if len(m) == 3}.values()}

    def get_name(self) -> str:
        return 'Nvidia'

    async def _assign_mode(self, enable: bool, user_env: Dict[str, str], context: Dict[str, str]) -> bool:
        mode = context.get('mode')

        if not mode:
            self._log.error(f"Cannot {'enable' if enable else 'disable'} the {self.get_name()} compositor: no mode on context")
            return False

        meta_mode = '{' + f"{mode}={'On' if enable else 'Off'}" + '}'
        cmd = f'{self._main_cmd} --assign CurrentMetaMode="nvidia-auto-select +0+0 {meta_mode}"'
        self._log.debug(f"{'Enabling' if enable else 'Disabling'} {self.get_name()} compositor: {cmd}")
        code, output = await system.async_syscall(cmd, custom_env=user_env)

        if code == 0 and (not output or 'error assigning value' not in output.lower()):
            return True

        log_err = '. Command output: {}'.format(output.replace('\n', ' ')) if output else ''
        self._log.error(f"Could not {'enable' if enable else 'disable'} {self.get_name()} compositor{log_err}")
        return False

    async def enable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._assign_mode(True, user_env, context)

    async def disable(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> bool:
        return await self._assign_mode(False, user_env, context)

    async def is_enabled(self, user_id: Optional[int], user_env: Optional[dict], context: dict) -> Optional[bool]:
        code, output = await system.async_syscall(f'{self._main_cmd} -q /CurrentMetaMode', custom_env=user_env)

        if code == 0:
            output = output.strip()

            if output:
                attrs = self.extract_attributes(output)

                if attrs:
                    context['mode'] = 'ForceFullCompositionPipeline' if len(attrs) == 2 else 'ForceCompositionPipeline'
                    return True

            if context.get('mode'):
                return False
            else:
                output_log = output.replace('\n', ' ')
                self._log.warning(f"Could not determine {self.get_name()} compositor state from 'nvidia-settings' output: {output_log}")
        else:
            output_log = '. Output: {}'.format(output.replace('\n', ' ')) if output else ''
            self._log.error(f"Error while checking {self.get_name()} compositor state. 'nvidia-settings' returned an unexpected exitcode ({code}){output_log}")


async def inxi_read_compositor(user_id: int, user_env: Optional[dict], logger: Logger) -> Optional[str]:
    if which('inxi'):
        cmd = 'inxi -Gxx -c 0'
        code, output = await run_async_user_process(cmd=cmd, user_id=user_id, user_env=user_env)

        if code == 0:
            name = RE_COMPOSITOR_NAME.findall(output)

            if not name:
                logger.warning(f"Command '{cmd}' did not return the window compositor name")
            else:
                return name[0].strip().lower()
        else:
            output_log = output.replace('\n', ' ') if output else ''
            logger.error(f"Error when executing command '{cmd}'. Could not read the current window compositor. Exit code: {code}. Output: {output_log}")


def guess_compositor_for_desktop_environment(user_env: Optional[dict], logger: Logger) -> Optional[str]:
    desk_env = user_env.get('XDG_CURRENT_DESKTOP', '').lower() if user_env else None

    if not desk_env:
        logger.warning("Could not determine the desktop environment and compositor (missing variable 'XDG_CURRENT_DESKTOP'")
        return

    logger.info(f"Guessing window compositor based on desktop environment: {desk_env}")

    if desk_env == 'kde':
        return 'kwin'
    elif desk_env == 'xfce':
        return 'xfwm4'
    elif desk_env == 'mate':
        return 'marco'
    else:
        logger.warning(f"Unknown window compositor for desktop environment: {desk_env}")


async def get_window_compositor(user_id: int, user_env: Optional[Dict[str, str]], logger: Logger) -> Optional[WindowCompositor]:
    name = await inxi_read_compositor(user_id, user_env, logger)

    if not name:
        name = guess_compositor_for_desktop_environment(user_env, logger)

    return get_window_compositor_by_name(name, logger)


def get_window_compositor_by_name(name: str, logger: Logger) -> Optional[WindowCompositor]:
    if name:
        clean_name = name.strip().lower()

        if 'kwin' in clean_name:
            return KWinCompositor(logger)
        elif 'xfwm4' in clean_name:
            return Xfwm4Compositor(logger)
        elif 'marco' in clean_name or 'metacity' in clean_name:
            return MarcoCompositor(logger)
        elif 'compton' in clean_name:
            return PicomCompositor('compton', logger)
        elif 'picom' in clean_name:
            return PicomCompositor('picom', logger)
        elif 'compiz' in clean_name:
            return CompizCompositor(logger)
        elif 'nvidia' == clean_name:
            return NvidiaCompositor(logger)
        else:
            logger.warning(f"Compositor '{name}' is currently not supported")



import os
import pwd

from typing import Dict, Optional


def map_home_users() -> Dict[str, int]:
    return {p.pw_name: p.pw_uid for p in pwd.getpwall() if (p.pw_name == 'root' or p.pw_dir.startswith('/home/'))}


def map_all_users() -> Dict[str, int]:
    return {p.pw_name: p.pw_uid for p in pwd.getpwall()}


def is_root_user(uid: Optional[int] = None) -> bool:
    return os.getuid() == 0 if uid is None else uid == 0



from typing import Dict

from guapow.cli.commands.optimizer import *
from guapow.cli.commands.profile import *
from guapow.cli.commands.watcher import *


def map_commands(logger: Logger) -> Dict[str, CLICommand]:
    res = {}

    for c in CLICommand.__subclasses__():
        instance = c(logger)
        res[instance.get_command()] = instance

    return res



from asyncio import Lock
from typing import Set, Optional


class OptimizationQueue:
    """
    Responsible to properly control what is being processed
    """

    def __init__(self, queued_pids: Set[int]):
        self._queued_pids = queued_pids
        self._lock_queued_pids = Lock()

    async def add_pid(self, pid: int) -> bool:
        async with self._lock_queued_pids:
            if pid not in self._queued_pids:
                self._queued_pids.add(pid)
                return True

        return False

    async def remove_pids(self, *pids: int):
        if pids is not None:
            async with self._lock_queued_pids:
                for pid in pids:
                    self._queued_pids.discard(pid)

    def get_view(self) -> Optional[Set[int]]:
        if self._queued_pids is not None:
            return {*self._queued_pids}

    @classmethod
    def empty(cls) -> "OptimizationQueue":
        return OptimizationQueue(set())



from argparse import ArgumentParser, Namespace
from typing import Iterable

from guapow import __app_name__, __version__
from guapow.cli.command import CLICommand


def read(commands: Iterable[CLICommand]) -> Namespace:
    parser = ArgumentParser(prog=__app_name__, description=f"Utility commands for {__app_name__}")
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s {__version__}')

    sub_parsers = parser.add_subparsers(dest='command', help='Available commands')

    for c in commands:
        c.add(sub_parsers)

    return parser.parse_args()



import os
from abc import ABC
from asyncio import Lock
from typing import Optional, Tuple, List

from guapow.common.scripts import RunScripts
from guapow.common.users import is_root_user
from guapow.runner.profile import RunnerProfile
from guapow.runner.task import StopProcesses, RunnerContext
from guapow.service.optimizer import cpu
from guapow.service.optimizer.cpu import GOVERNOR_FILE_PATTERN
from guapow.service.optimizer.gpu import GPUDriver
from guapow.service.optimizer.task.model import Task, OptimizationContext, OptimizedProcess, CPUState
from guapow.service.optimizer.win_compositor import get_window_compositor


class EnvironmentTask(Task, ABC):
    """
    Task not related to the optimized process
    """
    pass


class ChangeCPUFrequencyGovernor(EnvironmentTask):

    def __init__(self, context: OptimizationContext, cpu0_governor_file: Optional[str] = GOVERNOR_FILE_PATTERN.format(0)):
        super(ChangeCPUFrequencyGovernor, self).__init__(context)
        self._cpufreq_man = context.cpufreq_man
        self._cpu0_governor_file = cpu0_governor_file
        self._cpu_count = context.cpu_count

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        if self._cpu_count == 0:
            return False, "No CPU detected"

        if os.path.exists(self._cpu0_governor_file):
            if is_root_user():
                return True, None
            else:
                return False, "It will not be possible to change the CPUs scaling governors: requires root privileges"

        return False, f"It will not be possible to change the CPUs scaling governors: file '{self._cpu0_governor_file}' not found"

    async def should_run(self, process: OptimizedProcess) -> bool:
        return bool(process.profile.cpu and process.profile.cpu.performance)

    def is_allowed_for_self_requests(self) -> bool:
        return True

    async def run(self, process: OptimizedProcess):
        async with self._cpufreq_man.lock():
            current_governors = await self._cpufreq_man.map_current_governors()
            prev_governors = {}

            for gov, cpus in current_governors.items():
                if gov != cpu.GOVERNOR_PERFORMANCE:
                    changed_cpus = await self._cpufreq_man.change_governor(cpu.GOVERNOR_PERFORMANCE, cpus)
                    if changed_cpus:
                        prev_governors[gov] = changed_cpus

            if not process.request.is_self_request:
                if prev_governors:
                    self._cpufreq_man.save_governors(prev_governors)
                    process.previous_cpu_state = CPUState(prev_governors)
                else:
                    saved_governors = self._cpufreq_man.get_saved_governors()

                    if saved_governors:
                        process.previous_cpu_state = CPUState(saved_governors)


class ChangeGPUModeToPerformance(EnvironmentTask):

    def __init__(self, context: OptimizationContext):
        super(ChangeGPUModeToPerformance, self).__init__(context=context)
        self._context = context
        self._log = context.logger
        self._gpu_man = context.gpu_man

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        if self._context.system_service or not self._gpu_man.is_cache_enabled():
            return True, None
        elif await self._list_drivers_with_gpus():
            return True, None
        else:
            return False, "No manageable GPUs found"

    async def _list_drivers_with_gpus(self) -> List[GPUDriver]:
        self._log.debug("Checking available GPUs")
        return [gpu async for gpu, _ in self._gpu_man.map_working_drivers_and_gpus()]

    async def should_run(self, process: OptimizedProcess) -> bool:
        return bool(process.profile.gpu and process.profile.gpu.performance)

    async def run(self, process: OptimizedProcess):
        target_gpu_ids = self._context.gpu_ids if self._context.gpu_ids else None
        previous_gpu_states = await self._gpu_man.activate_performance(user_environment=process.request.user_env,
                                                                       target_gpu_ids=target_gpu_ids)

        if previous_gpu_states:
            process.previous_gpus_states = previous_gpu_states


class DisableWindowCompositor(EnvironmentTask):

    def __init__(self, context: OptimizationContext, available: Optional[bool] = None):
        super(DisableWindowCompositor, self).__init__(context=context)
        self._log = context.logger
        self._context = context
        self._available = available
        self._compositor_checked = False
        self._manageable: Optional[bool] = None
        self._lock = Lock()

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    async def should_run(self, process: OptimizedProcess) -> bool:
        request, profile = process.request, process.profile

        if profile.compositor and profile.compositor.off:
            async with self._lock:
                if not self._context.compositor and not self._compositor_checked:
                    compositor = await get_window_compositor(logger=self._log, user_id=request.user_id, user_env=request.user_env)

                    if compositor:
                        self._log.info(f"Window compositor detected: {compositor.get_name()}")
                        self._context.compositor = compositor

                    self._compositor_checked = True

                if self._context.compositor and self._manageable is None:
                    res, msg = self._context.compositor.can_be_managed()
                    self._manageable = res

                    if not self._manageable:
                        log_warn = f'. Reason: {msg}' if msg else ''
                        self._log.warning(f"Compositor {self._context.compositor.get_name()} cannot be managed{log_warn}")

            return bool(self._context.compositor and self._manageable)

        return False

    async def run(self, process: OptimizedProcess):
        async with self._context.compositor.lock():
            context = {}
            enabled = await self._context.compositor.is_enabled(user_id=process.user_id, user_env=process.user_env, context=context)

            if enabled is None:
                self._log.error(f"It will not be possible to disable the window compositor for process '{process.pid}'")
                return

            if not enabled:
                self._log.info("Window compositor is already disabled")
                return

            if await self._context.compositor.disable(user_id=process.user_id, user_env=process.user_env, context=context):
                self._log.info("Window compositor disabled")
                self._context.compositor_disabled_context = context


class HideMouseCursor(EnvironmentTask):

    def __init__(self, context: OptimizationContext):
        self._context = context
        self._log = context.logger
        self._mouse_man = context.mouse_man

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return self._mouse_man.can_work()

    async def should_run(self, process: OptimizedProcess) -> bool:
        return bool(process.profile.hide_mouse)

    async def run(self, process: OptimizedProcess):
        await self._mouse_man.hide_cursor(user_request=not process.request.is_self_request, user_env=process.request.user_env)


class StopProcessesAfterLaunch(EnvironmentTask):  # act as a wrapper for 'runner.task.StopProcesses'

    def __init__(self, context: OptimizationContext):
        self._log = context.logger
        self._context = context
        self._task_context = RunnerContext(environment_variables=None, processes_initialized=None,
                                           logger=self._log, stopped_processes=None)
        self._task = StopProcesses(self._task_context)

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return self._task.is_available()

    async def should_run(self, process: OptimizedProcess) -> bool:
        return process.profile.stop_after and process.profile.stop_after.processes

    async def run(self, process: OptimizedProcess):
        self._task_context.stopped_processes = {}

        fake_profile = RunnerProfile(path=None, environment_variables=None)
        fake_profile.stop = process.profile.stop_after
        await self._task.run(fake_profile)

        if self._task_context.stopped_processes:
            process.stopped_after_launch = self._task_context.stopped_processes


class RunPostLaunchScripts(EnvironmentTask):

    def __init__(self, context: OptimizationContext):
        self._context = context
        self._task = RunScripts(name='post launch', root_allowed=context.allow_root_scripts, logger=context.logger)

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    async def should_run(self, process: OptimizedProcess) -> bool:
        return bool(process.profile.after_scripts and process.profile.after_scripts.scripts)

    async def run(self, process: OptimizedProcess):
        started_pids = await self._task.run(scripts=[process.profile.after_scripts], user_id=process.user_id, user_env=process.user_env)

        if started_pids:
            process.related_pids.update(started_pids)


class ChangeCPUEnergyPolicyLevel(EnvironmentTask):

    def __init__(self, context: OptimizationContext):
        super(ChangeCPUEnergyPolicyLevel, self).__init__(context)
        self._man = context.cpuenergy_man
        self._log = context.logger

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return self._man.can_work()

    async def should_run(self, process: OptimizedProcess) -> bool:
        return bool(process.profile.cpu and process.profile.cpu.performance)

    def is_allowed_for_self_requests(self) -> bool:
        return True

    async def run(self, process: OptimizedProcess):
        async with self._man.lock():
            current_cpu_states = await self._man.map_current_state()

            if not current_cpu_states:
                self._log.error("Could not determine the current CPUs energy policy level")
                return

            not_in_performance = {i: state for i, state in current_cpu_states.items()
                                  if state != self._man.LEVEL_PERFORMANCE}

            if not not_in_performance:
                process.cpu_energy_policy_changed = bool(self._man.saved_state)
                return

            cpus_changed_state = await self._man.change_states({i: self._man.LEVEL_PERFORMANCE
                                                                for i in not_in_performance})
            cpus_changed, cpus_not_changed = [], []

            for idx, changed in cpus_changed_state.items():
                if changed:
                    cpus_changed.append(idx)
                else:
                    cpus_not_changed.append(idx)

            if cpus_not_changed:
                self._log.error(f"Could not change the energy policy level to full performance "
                                f"({self._man.LEVEL_PERFORMANCE}) for the following CPUs: "
                                f"{', '.join(str(i) for i in sorted(cpus_not_changed))}")

            if cpus_changed:
                cpus_changed.sort()
                self._log.info(f"Energy policy level changed to full performance ({self._man.LEVEL_PERFORMANCE}) "
                               f"for the following CPUs: {', '.join(str(i) for i in sorted(cpus_changed))}")

                if not process.request.is_self_request:
                    state_to_save = {idx: current_cpu_states[idx] for idx in cpus_changed}
                    self._log.debug(f"Previous CPUs energy policy levels state saved: {state_to_save}")
                    self._man.save_state(state_to_save)
                    process.cpu_energy_policy_changed = True



from typing import Optional

import aiofiles


def get_machine_id_path() -> str:
    return '/etc/machine-id'


async def read_machine_id() -> Optional[str]:
    try:
        async with aiofiles.open(get_machine_id_path()) as f:
            return (await f.read()).strip()
    except FileNotFoundError:
        return



from logging import Logger
from typing import Optional, Dict, Tuple

import aiofiles

from guapow.common.model import ProfileFile, ScriptSettings
from guapow.common.model_util import FileModelFiller
from guapow.common.profile import StopProcessSettings
from guapow.common.profile import get_profile_dir, get_default_profile_name


class RunnerProfile(ProfileFile):

    FILE_MAPPING = {'proc.env': ('environment_variables', dict, None)}

    def __init__(self, path: Optional[str], environment_variables: Optional[dict]):
        super(RunnerProfile, self).__init__(path)
        self.before_scripts = ScriptSettings(node_name='scripts.before')
        self.environment_variables = environment_variables
        self.stop = StopProcessSettings('stop.before', None)

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def is_valid(self) -> bool:
        if super(RunnerProfile, self).is_valid():
            return True

        return bool(self.environment_variables)

    def setup_valid_properties(self):
        if self.before_scripts:
            self.before_scripts.run_as_root = False

    @classmethod
    def empty(cls, path: Optional[str] = None) -> "RunnerProfile":
        return cls(path, None)


class RunnerProfileReader:

    def __init__(self, model_filler: FileModelFiller, logger: Logger):
        self._model_filler = model_filler
        self._log = logger

    async def read_valid(self, user_id: int, user_name: str, profile: str, add_settings: Optional[str] = None) -> Optional[RunnerProfile]:
        profile_path = f'{get_profile_dir(user_id, user_name)}/{profile}.profile'
        self._log.info(f"Reading runner profile file '{profile_path}'")

        async with aiofiles.open(profile_path) as f:
            profile_str = (await f.read()).strip()

        if not profile_str:
            self._log.warning(f"No properties defined in runner profile file '{profile_path}'")
            return

        instance = RunnerProfile.empty()
        self._model_filler.fill_profile(profile=instance, profile_str=profile_str, profile_path=profile_path, add_settings=add_settings)
        instance.setup_valid_properties()

        if instance.is_valid():
            return instance

        self._log.warning(f"No valid runner profile properties found in file '{profile_path}'")

    async def read_available(self, user_id: int, user_name: str, profile: Optional[str], add_settings: Optional[str] = None) -> Optional[RunnerProfile]:
        if profile:
            try:
                return await self.read_valid(user_id=user_id, user_name=user_name, profile=profile, add_settings=add_settings)
            except FileNotFoundError as e:
                self._log.warning(f"Runner profile file '{e.filename}' not found")

        try:
            return await self.read_valid(user_id=user_id, user_name=user_name, profile=get_default_profile_name(), add_settings=add_settings)
        except FileNotFoundError as e:
            self._log.warning(f"Runner profile file '{e.filename}' not found")

    def map_valid_config(self, config: str) -> Optional[RunnerProfile]:
        instance = RunnerProfile.empty()
        self._model_filler.fill_profile(profile=instance, profile_str='\n'.join(config.split(' ')), profile_path=None)
        instance.reset_invalid_nested_members()
        instance.setup_valid_properties()

        if instance.is_valid():
            return instance
        else:
            self._log.error(f'Invalid configuration defined: {config}')



import asyncio
import glob
import os
import time
from logging import Logger
from typing import Optional, Dict, List, Tuple

import aiofiles

from guapow.common.config import OptimizerConfig
from guapow.common.model import FileModel, CustomEnum, ProfileFile, ScriptSettings
from guapow.common.model_util import FileModelFiller
from guapow.common.profile import StopProcessSettings, get_profile_dir


class CPUSchedulingPolicy(CustomEnum):
    RR = os.SCHED_RR, True, True  # round-robin
    FIFO = os.SCHED_FIFO, True, True
    BATCH = os.SCHED_BATCH, False, False
    OTHER = os.SCHED_OTHER, False, False
    IDLE = os.SCHED_IDLE, False, False

    def __init__(self, value: str, priority: bool, root: bool):
        self._value = value
        self._priority = priority
        self._root = root

    def requires_priority(self) -> bool:
        return self._priority

    def requires_root(self) -> bool:
        return self._root

    def value(self):
        return self._value


class IOSchedulingClass(CustomEnum):
    NONE = '0', False
    REALTIME = '1', True
    BEST_EFFORT = '2', True
    IDLE = '3', False

    def __init__(self, value: int, priority: bool):
        self._value = value
        self.priority = priority

    def supports_priority(self) -> bool:
        return self.priority

    def value(self):
        return self._value

    def __str__(self) -> str:
        return self.name.lower()


class ProcessSchedulingSettings(FileModel):
    FILE_MAPPING = {'policy': ('policy', CPUSchedulingPolicy, None),
                    'policy.priority': ('priority', int, None)}

    def __init__(self, policy: Optional[CPUSchedulingPolicy], policy_priority: Optional[int] = None):
        self.policy = policy
        self.priority = policy_priority

    def is_valid(self) -> bool:
        return bool(self.policy)

    def has_valid_priority(self) -> bool:
        if not self.policy.requires_priority():
            return False

        return self.priority is not None and 0 < self.priority < 100

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        pass


class ProcessNiceSettings(FileModel):

    FILE_MAPPING = {'nice': ('level', int, None),
                    'nice.delay': ('delay', float, None),
                    'nice.watch': ('watch', bool, True)}

    def __init__(self, nice_level: Optional[int], delay: Optional[float], watch: Optional[bool]):
        self.delay = delay
        self.level = nice_level
        self.watch = watch

    def has_valid_level(self) -> bool:
        return self.level is not None and -21 < self.level < 20

    def get_output_name(self) -> str:
        return ""

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        pass

    def is_valid(self) -> bool:
        return self.has_valid_level()


class ProcessSettings(FileModel):

    MAPPING = {'affinity': ('cpu_affinity', List[int], None)}

    def __init__(self, affinity: Optional[List[int]]):
        self.scheduling = ProcessSchedulingSettings(None, None)
        self.io = IOScheduling(None, None)
        self.nice = ProcessNiceSettings(None, None, None)
        self.cpu_affinity = affinity

    def has_valid_cpu_affinity(self, cpu_count: int) -> bool:
        if cpu_count <= 0:
            return False

        if not self.cpu_affinity:
            return False

        for idx in self.cpu_affinity:
            if idx < 0 or idx >= cpu_count:
                return False

        return True

    def is_valid(self) -> bool:
        if self.cpu_affinity:
            return True

        valid = False
        for v in self.__dict__.values():
            if v is not None and isinstance(v, FileModel):
                if v.is_valid():
                    valid = True

        return valid

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        return 'proc'

    @classmethod
    def empty(cls) -> "ProcessSettings":
        instance = ProcessSettings(None)
        instance.scheduling = None
        instance.nice = None
        instance.io = None
        return instance


class CPUSettings(FileModel):
    FILE_MAPPING = {'performance': ('performance', bool, True)}

    def __init__(self, performance: Optional[bool]):
        self.performance = performance

    def is_valid(self) -> bool:
        return self.performance is not None

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        return 'cpu'


class IOScheduling(FileModel):

    FILE_MAPPING = {'class': ('ioclass', IOSchedulingClass, None), 'nice': ('nice_level', int, None)}

    def __init__(self, ioclass: Optional[IOSchedulingClass], nice_level: Optional[int]):
        self.ioclass = ioclass
        self.nice_level = nice_level

    def is_valid(self) -> bool:
        return bool(self.ioclass)

    def has_valid_priority(self) -> bool:
        return self.nice_level is not None and 0 <= self.nice_level < 8

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        return 'io'


class GPUSettings(FileModel):

    FILE_MAPPING = {'performance': ('performance', bool, True)}

    def __init__(self, performance: Optional[bool]):
        self.performance = performance

    def is_valid(self) -> bool:
        return self.performance is not None

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        return 'gpu'


class CompositorSettings(FileModel):

    FILE_MAPPING = {'off': ('off', bool, True)}

    def __init__(self, off: Optional[bool]):
        self.off = off

    def is_valid(self) -> bool:
        return self.off is True

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        return 'compositor'


class LauncherSettings(FileModel):
    FILE_MAPPING = {'launcher': ('mapping', dict, None),
                    'launcher.skip_mapping': ('skip_mapping', bool, True)}

    def __init__(self, mapping: Optional[dict], skip_mapping: Optional[bool]):
        self.mapping = mapping
        self.skip_mapping = skip_mapping

    def is_valid(self) -> bool:
        return bool(self.mapping) or self.skip_mapping is not None

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.FILE_MAPPING

    def get_file_root_node_name(self) -> Optional[str]:
        pass


class OptimizationProfile(ProfileFile):

    MAPPING = {'mouse.hidden': ('hide_mouse', bool, True), 'steam': ('steam', bool, True)}

    def __init__(self, path: Optional[str], cpu: Optional[CPUSettings], steam: Optional[bool],
                 gpu: Optional[GPUSettings], process: Optional[ProcessSettings], compositor: Optional[CompositorSettings],
                 launcher: Optional[LauncherSettings], hide_mouse: Optional[bool]):
        super(OptimizationProfile, self).__init__(path)
        self.cpu = cpu
        self.steam = steam
        self.gpu = gpu
        self.process = process
        self.after_scripts = ScriptSettings('scripts.after')
        self.finish_scripts = ScriptSettings('scripts.finish')
        self.compositor = compositor
        self.launcher = launcher
        self.hide_mouse = hide_mouse
        self.stop_after = StopProcessSettings('stop.after', None)  # processes to stop after the process to be optimized launches

    def get_file_mapping(self) -> Dict[str, Tuple[str, type, Optional[object]]]:
        return self.MAPPING

    def get_output_name(self) -> str:
        return 'profile'

    def from_config(self) -> bool:
        return not self.path

    def get_log_str(self) -> str:
        return "informed configuration" if self.from_config() else f"profile '{self.name}'"

    def is_valid(self) -> bool:
        if super(OptimizationProfile, self).is_valid():
            return True

        return any([self.hide_mouse is not None,
                    self.steam is not None])

    @classmethod
    def empty(cls, path: Optional[str] = None) -> "OptimizationProfile":
        instance = cls(path=path, cpu=None, gpu=None, steam=None, process=None, compositor=None, launcher=None, hide_mouse=None)
        instance.finish_scripts = None
        instance.after_scripts = None
        instance.stop_after = None
        return instance

    @classmethod
    def raw(cls, path: Optional[str] = None) -> "OptimizationProfile":
        return cls(path=path,
                   cpu=CPUSettings(None),
                   steam=None,
                   process=ProcessSettings(None),
                   gpu=GPUSettings(None),
                   compositor=CompositorSettings(None),
                   launcher=LauncherSettings(None, None),
                   hide_mouse=None)

    @classmethod
    def from_optimizer_config(cls, config: OptimizerConfig) -> Optional["OptimizationProfile"]:
        if config and config.cpu_performance:
            profile = cls.empty(None)

            if config.cpu_performance:
                profile.cpu = CPUSettings(True)

            return profile


class OptimizationProfileCache:

    def __init__(self, logger: Logger):
        self._cache: Dict[str, OptimizationProfile] = dict()
        self._log = logger

    @staticmethod
    def map_key(path: str, add_settings: Optional[str]) -> str:
        return f"{path}{f'#{add_settings}' if add_settings else ''}"

    def _map_extra_log(self, add_settings: Optional[str]) -> str:
        return f" ({add_settings})" if add_settings else ''

    def get(self, path: str, add_settings: Optional[str] = None) -> Optional[OptimizationProfile]:
        ti = time.time()
        if self._cache:
            instance = self._cache.get(self.map_key(path, add_settings))
            if instance:
                tf = time.time()
                self._log.debug(f"Cached profile '{path}'{self._map_extra_log(add_settings)} found in {tf - ti:.5f} seconds")
                return instance

        self._log.debug(f"No cached profile found for '{path}'{self._map_extra_log(add_settings)}")

    def add(self, path: str,  profile: OptimizationProfile, add_settings: Optional[str]):
        if profile:
            self._log.debug(f"Caching profile '{path}'{self._map_extra_log(add_settings)}")
            self._cache[self.map_key(path, add_settings)] = profile

    @property
    def size(self) -> int:
        return len(self._cache)


class OptimizationProfileReader:

    def __init__(self, model_filler: FileModelFiller, logger: Logger, cache: Optional[OptimizationProfileCache]):
        self._model_filler = model_filler
        self._log = logger
        self._cache = cache

    def map(self, profile_str: str, profile_path: Optional[str] = None, add_settings: Optional[str] = None) -> OptimizationProfile:

        profile = OptimizationProfile.raw()

        self._model_filler.fill_profile(profile=profile, profile_str=profile_str,
                                        profile_path=profile_path, add_settings=add_settings)
        return profile

    async def read(self, profile_path: str, add_settings: Optional[str] = None) -> Optional[OptimizationProfile]:
        ti = time.time()

        async with aiofiles.open(profile_path) as f:
            profile_str = await f.read()

        profile_str = profile_str.strip()

        if not profile_str:
            self._log.warning(f"No properties defined in profile file '{profile_path}'")
            return

        instance = self.map(profile_str=profile_str, profile_path=profile_path, add_settings=add_settings)
        tf = time.time()
        self._log.debug(f"Profile file '{profile_path}' read and mapped in {tf - ti:.5f} seconds")
        return instance

    async def read_valid(self, profile_path: str, add_settings: Optional[str] = None, handle_not_found: bool = True) -> Optional[OptimizationProfile]:
        if self._cache:
            profile = self._cache.get(profile_path, add_settings)

            if profile:
                return profile

        try:
            profile = await self.read(profile_path=profile_path, add_settings=add_settings)
        except FileNotFoundError:
            if handle_not_found:
                profile = None
                self._log.warning(f"Profile file '{profile_path}' not found")
            else:
                raise

        if profile:
            if profile.is_valid():
                if self._cache:
                    self._cache.add(profile_path, profile, add_settings)

                return profile
            else:
                self._log.warning(f"Invalid profile file '{profile_path}'")

    @property
    def cached_profiles(self) -> int:
        return self._cache.size if self._cache else 0


async def cache_profiles(reader: OptimizationProfileReader, logger: Logger):
    profile_paths = set()
    for exp in {f'{d}/*.profile' for d in (get_profile_dir(0, 'root'), get_profile_dir(1, '*'))}:
        profile_paths.update(glob.glob(exp))

    if profile_paths:
        logger.debug(f"{len(profile_paths)} profiles found on disk")
        await asyncio.gather(*(reader.read_valid(path) for path in profile_paths))
        logger.info(f"{reader.cached_profiles} valid profiles cached")
    else:
        logger.info("No profile file found on disk to cache")



from abc import ABC
from typing import Type, List, Optional, TypeVar

ANY_CLASS = TypeVar("ANY_CLASS")


def instantiate_subclasses(root: Type[ANY_CLASS]) -> Optional[List[ANY_CLASS]]:
    instances = []

    root_subs = root.__subclasses__()

    if root_subs:
        for sub in root_subs:
            if ABC not in sub.__bases__:
                instances.append(sub())

            children = instantiate_subclasses(sub)

            if children:
                instances.extend(children)

    return instances if instances else None



import os.path
import re
from typing import Optional, Tuple

RE_STEAM_CMD = re.compile(r'^.+\s+SteamLaunch\s+AppId\s*=\s*\d+\s+--\s+(.+)')
RE_PROTON_CMD = re.compile(r'^.+/proton\s+waitforexitandrun\s+(/.+)$')
RE_EXE_NAME = re.compile(r'^(.+\.\w+)(\s+.+)?$')


def get_exe_name(file_path: str) -> Optional[str]:
    exe_name = RE_EXE_NAME.findall(os.path.basename(file_path))
    return exe_name[0][0].strip() if exe_name else None


def get_proton_exec_name_and_paths(cmd: str) -> Optional[Tuple[Optional[str], str, str]]:
    if cmd:
        result = RE_PROTON_CMD.findall(cmd)
        if result:
            return get_exe_name(result[0]), 'Z:{}'.format(result[0].replace('/', '\\')), result[0]


def get_steam_runtime_command(cmd: str) -> Optional[str]:
    if cmd:
        result = RE_STEAM_CMD.findall(cmd)
        if result:
            return result[0].strip()



import os
from logging import Logger
from typing import Optional, Set, Tuple

import aiofiles

from guapow import __app_name__
from guapow.common.users import is_root_user

FILE_NAME = 'watch.ignore'


def get_file_path(user_id: int, user_name: str) -> Optional[str]:
    if is_root_user(user_id):
        return f'/etc/{__app_name__}/{FILE_NAME}'
    else:
        return f'/home/{user_name}/.config/{__app_name__}/{FILE_NAME}'


def get_existing_file_path(user_id: int, user_name: str, logger: Logger) -> Optional[str]:
    file_path = get_file_path(user_id, user_name)
    if os.path.isfile(file_path):
        logger.info(f"Ignored file '{file_path}' found")
        return file_path
    else:
        logger.warning(f"Ignored file '{file_path}' not found")
        return get_existing_file_path(0, 'root', logger) if not is_root_user(user_id) else None


def get_default_file_path(user_id: int, user_name: str, logger: Logger) -> str:
    file_path = get_existing_file_path(user_id, user_name, logger)

    if file_path:
        return file_path

    default_path = get_file_path(user_id, user_name)
    logger.info(f'Considering the default ignored file path for current user: {default_path}')
    return default_path


async def read(file_path: str, logger: Logger, last_file_found_log: Optional[bool]) -> Tuple[bool, Optional[Set[str]]]:
    try:
        async with aiofiles.open(file_path) as f:
            ignored_str = (await f.read()).strip()

        if not last_file_found_log:
            logger.info(f"Ignore file '{file_path}' found")

    except FileNotFoundError:
        if last_file_found_log is None or last_file_found_log is True:
            logger.debug(f"Ignore file '{file_path}' not found")

        return False, None

    ignored = set()

    for string in ignored_str.split('\n'):
        clean_string = string.strip()

        if clean_string:
            sharp_less = clean_string.split('#')
            final_str = sharp_less[0].strip()

            if final_str:
                ignored.add(final_str)

    return True, ignored if ignored else None



from abc import ABC, abstractmethod
from typing import List, Optional, Dict, Set, Type, Tuple

from guapow.common.model import ScriptSettings
from guapow.service.optimizer.gpu import GPUState, GPUDriver
from guapow.service.optimizer.post_process.summary import PostProcessSummary
from guapow.service.optimizer.task.model import CPUState


class PostProcessContext:

    def __init__(self, restorable_cpus: Optional[List[CPUState]],
                 restorable_gpus: Optional[Dict[Type[GPUDriver], List[GPUState]]],
                 pids_to_stop: Optional[Set[int]],
                 scripts: Optional[List[ScriptSettings]],
                 user_id: Optional[int],
                 user_env: Optional[dict],
                 restore_compositor: Optional[bool],
                 stopped_processes: Optional[List[Tuple[str, Optional[str]]]],
                 not_stopped_processes: Optional[Set[str]],
                 restore_mouse_cursor: Optional[bool],
                 restore_cpu_energy_policy: Optional[bool]):
        self.restorable_cpus = restorable_cpus
        self.restorable_gpus = restorable_gpus
        self.pids_to_stop = pids_to_stop
        self.scripts = scripts
        self.user_env = user_env
        self.user_id = user_id
        self.restore_compositor = restore_compositor
        self.stopped_processes = stopped_processes
        self.not_stopped_processes = not_stopped_processes
        self.restore_mouse_cursor = restore_mouse_cursor
        self.restore_cpu_energy_policy = restore_cpu_energy_policy

    @classmethod
    def empty(cls):
        return cls(restorable_cpus=None, restorable_gpus=None, pids_to_stop=None,
                   scripts=None, user_env=None, user_id=None, restore_compositor=None,
                   stopped_processes=None, restore_mouse_cursor=None, not_stopped_processes=None,
                   restore_cpu_energy_policy=None)


class PostContextFiller(ABC):

    @abstractmethod
    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        pass


class RestorableCPUGovernorsFiller(PostContextFiller):

    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        if not summary.cpus_in_use and summary.previous_cpus_states:
            context.restorable_cpus = summary.previous_cpus_states


class RestorableCPUEnergyPolicyLevelFiller(PostContextFiller):

    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        context.restore_cpu_energy_policy = not summary.keep_cpu_energy_policy and summary.restore_cpu_energy_policy


class RestorableGPUsFiller(PostContextFiller):

    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        if summary.previous_gpus_states:
            gpus = summary.previous_gpus_states

            if summary.gpus_in_use:
                gpus_in_use = summary.gpus_in_use

                drivers_to_restore = {*gpus.keys()}
                for driver in drivers_to_restore:
                    ids = gpus_in_use.get(driver)

                    if ids:
                        to_restore = [s for s in gpus[driver] if s.id not in ids]

                        if to_restore:
                            gpus[driver] = to_restore
                        else:
                            del gpus[driver]

            if gpus:
                context.restorable_gpus = gpus


class SortedFinishScriptsFiller(PostContextFiller):

    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        if summary.post_scripts:
            sorted_scripts = []

            for _, scripts in sorted(summary.post_scripts.items()):  # sorted by optimization timestamp
                sorted_scripts.append(scripts)

            context.scripts = sorted_scripts


class SortedProcessesToRelaunchFiller(PostContextFiller):

    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        if summary.processes_relaunch_by_time:
            sorted_procs, unique_procs, stopped_names, not_stopped_names = [], set(), set(), set()

            for _, comm_cmd in sorted(summary.processes_relaunch_by_time.items()):  # sorted by optimization timestamp
                for comm, cmd in comm_cmd.items():
                    if not summary.processes_not_relaunch or comm not in summary.processes_not_relaunch:
                        if not cmd:
                            not_stopped_names.add(comm)
                        else:
                            proc = (comm, cmd)

                            if proc not in unique_procs:
                                stopped_names.add(comm)
                                unique_procs.add(proc)
                                sorted_procs.append(proc)

            if sorted_procs:
                context.stopped_processes = sorted_procs

            if not_stopped_names:
                actually_not_stopped = not_stopped_names.difference(stopped_names)

                if actually_not_stopped:
                    context.not_stopped_processes = actually_not_stopped


class PostProcessContextMapper(PostContextFiller):

    __instance: Optional["PostProcessContextMapper"] = None

    def __init__(self, fillers: Optional[List[PostContextFiller]]):
        self._fillers = fillers

    @classmethod
    def instance(cls) -> "PostProcessContextMapper":
        if cls.__instance is None:
            fillers = [sub() for sub in PostContextFiller.__subclasses__() if sub != cls]
            cls.__instance = cls(fillers)

        return cls.__instance

    def fill(self, context: PostProcessContext, summary: PostProcessSummary):
        if self._fillers:
            for filler in self._fillers:
                filler.fill(context, summary)

    def get_fillers(self) -> Optional[List[PostContextFiller]]:
        return [*self._fillers] if self._fillers else None

    def map(self, summary: PostProcessSummary) -> PostProcessContext:
        context = PostProcessContext.empty()

        if summary.user_id is not None:
            context.user_id = summary.user_id

        if summary.user_env:
            context.user_env = summary.user_env

        if summary.pids_to_stop:
            context.pids_to_stop = summary.pids_to_stop

        context.restore_compositor = not summary.keep_compositor_disabled and summary.restore_compositor
        context.restore_mouse_cursor = not summary.keep_mouse_hidden and summary.restore_mouse_cursor

        self.fill(context, summary)
        return context



import asyncio
import logging
import logging.handlers
import os
from datetime import datetime
from io import StringIO
from pathlib import Path
from queue import SimpleQueue as Queue

from guapow import __app_name__


class LocalQueueHandler(logging.handlers.QueueHandler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            self.enqueue(record)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.handleError(record)


class FilePathFilter(logging.Filter):

    def filter(self, record):
        record.module_path = record.pathname.split('site-packages/')[1] if 'site-packages' in record.pathname else str(record.pathname)
        return True


def _setup_logger_as_async(logger: logging.Logger):
    queue = Queue()
    logger.addHandler(LocalQueueHandler(queue))

    handlers = []
    for h in logger.handlers[:]:
        if not isinstance(h, LocalQueueHandler):
            handlers.append(h)
            logger.removeHandler(h)

    listener = logging.handlers.QueueListener(
        queue, *handlers, respect_handler_level=True
    )
    listener.start()


def get_log_format(level: int, service: bool) -> str:
    log_format = StringIO()

    if not service:
        log_format.write('%(asctime)s ')

    if level == logging.DEBUG:
        log_format.write('[%(module_path)s:%(lineno)s] ')

    log_format.write('%(message)s')
    log_format.seek(0)
    return log_format.read()


def new_logger(name: str, service: bool, enabled: bool, write_to_file: bool, threaded: bool = True, level: int = logging.INFO) -> logging.Logger:
    instance = logging.Logger(name, level=level)

    instance.addFilter(FilePathFilter())

    if enabled and write_to_file:
        log_dir = f'{Path.home()}/.local/share/{__app_name__}/log'
        Path(log_dir).mkdir(parents=True, exist_ok=True)
        now = datetime.now()
        handler = logging.FileHandler(f"{log_dir}/{now.strftime('%Y-%m-%d')}-{int(now.timestamp() * 1000000)}.log")
    else:
        handler = logging.StreamHandler()

    handler.setFormatter(logging.Formatter(get_log_format(level, service)))
    instance.addHandler(handler)
    instance.disabled = not enabled

    if threaded:
        _setup_logger_as_async(instance)

    return instance


def get_log_level(var_name: str) -> int:
    try:
        level = os.getenv(var_name, 'INFO').upper().strip()

        if level and hasattr(logging, level):
            return getattr(logging, level)
        else:
            return logging.INFO
    except ValueError:
        return logging.INFO



import asyncio
import re
import traceback
from abc import ABC, abstractmethod
from multiprocessing import Process
from re import Pattern
from typing import List, Optional, Dict, Set, Tuple, Type, Awaitable

from guapow.common import system
from guapow.common.scripts import RunScripts
from guapow.common.users import is_root_user
from guapow.service.optimizer.gpu import GPUState, GPUDriver
from guapow.service.optimizer.post_process.context import PostProcessContext
from guapow.service.optimizer.task.model import OptimizationContext, CPUState


class PostProcessTask(ABC):

    @abstractmethod
    def __init__(self, context: OptimizationContext):
        pass

    @abstractmethod
    def should_run(self, context: PostProcessContext) -> bool:
        pass

    @abstractmethod
    async def run(self, context: PostProcessContext):
        pass


class RestoreGPUState(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        super(RestoreGPUState, self).__init__(context)
        self._gpu_man = context.gpu_man
        self._log = context.logger

    def should_run(self, context: PostProcessContext) -> bool:
        return bool(context.restorable_gpus)

    async def _restore(self, driver: GPUDriver, states: List[GPUState], user_env: Optional[dict]):
        async with driver.lock():
            gpus_states = {}  # hold all states mapped to the same GPU

            for state in states:
                modes = gpus_states.get(state.id, set())
                gpus_states[state.id] = modes
                modes.add(state.power_mode)

            gpus_current_modes = await driver.get_power_mode({*gpus_states.keys()}, user_env)
            gpus_to_restore = {}

            for id_, modes in gpus_states.items():
                # if there is more than one mode mapped to same GPU, a default mode is preferred
                mode = [*modes][0] if len(modes) == 1 else driver.get_default_mode()
                current_mode = gpus_current_modes.get(id_)

                if mode:
                    if mode != current_mode:
                        gpus_to_restore[id_] = mode
                    else:
                        self._log.info(f"It is not necessary to restore {driver.get_vendor_name()} GPU ({id_}) to "
                                       f"'{mode.name.lower()}' mode")
                else:
                    self._log.error(f"Current mode unknown for {driver.get_vendor_name()} GPU '{id_}'")

            if gpus_to_restore:
                self._log.debug(f"Restoring power mode of {driver.get_vendor_name()} GPUS: "
                                f"{', '.join(gpus_to_restore)}")

                gpus_changed = await driver.set_power_mode(gpus_to_restore, user_env)

                if gpus_changed:
                    if not self._log.disabled:
                        not_restored = {gpu for gpu, changed in gpus_changed.items() if not changed}

                        if not_restored:
                            self._log.error(f"Could not restore power mode of {driver.get_vendor_name()}  GPUS: "
                                            f"{', '.join(gpus_changed)}")

                else:
                    self._log.error(f"Could not restore power mode of {driver.get_vendor_name()} GPUs: "
                                    f"{', '.join(gpus_to_restore.keys())}")

    async def run(self, context: PostProcessContext):
        restore_tasks = []
        for driver in self._gpu_man.get_drivers():
            states = context.restorable_gpus.get(driver.__class__)

            if states:
                restore_tasks.append(self._restore(driver, states, context.user_env))

        if restore_tasks:
            await asyncio.gather(*restore_tasks)


class RestoreCPUGovernor(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._cpufreq_man = context.cpufreq_man
        self._log = context.logger

    def should_run(self, context: PostProcessContext) -> bool:
        return bool(context.restorable_cpus)

    def _map_governors(self, governors: List[Dict[str, Set[int]]]) -> Tuple[Dict[str, Set[int]], Dict[int, Set[str]]]:
        governor_cpus, cpu_governors = {}, {}

        if governors:
            for govs in governors:
                if govs:  # it is possible that previous governors could not be determined because they were set to 'performance' at that time
                    for gov, cpus in govs.items():
                        gov_cpus = governor_cpus.get(gov, set())
                        gov_cpus.update(cpus)
                        governor_cpus[gov] = gov_cpus

                        for cpu in cpus:
                            govs = cpu_governors.get(cpu, set())
                            govs.add(gov)
                            cpu_governors[cpu] = govs

        return governor_cpus, cpu_governors

    def map_governors(self, cpu_states: List[CPUState]) -> Tuple[Dict[str, Set[int]], Dict[int, Set[str]]]:
        governor_cpus, cpu_governors = self._map_governors([state.governors for state in cpu_states])

        if not governor_cpus:
            governor_cpus, cpu_governors = self._map_governors([self._cpufreq_man.get_saved_governors()])

        return governor_cpus, cpu_governors

    def _remove_duplicates(self, governor_cpus: Dict[str, Set[int]], cpu_governors: Dict[int, Set[str]]):
        """
        if there is a CPU mapped to several governors, remove it from the governors with less CPUs mapped
        """
        sorted_governors_cpus = [g for n, g in sorted([(len(c), g) for g, c in governor_cpus.items()], reverse=True)]

        to_remove = {}  # governor by CPUs to remove
        for cpu, governors in cpu_governors.items():
            if len(governors) > 1:
                pref_gov_idx, pref_gov = None, None
                for gov in governors:
                    gov_prio = sorted_governors_cpus.index(gov)

                    if pref_gov_idx is None or pref_gov_idx > gov_prio:
                        pref_gov_idx = gov_prio
                        pref_gov = gov

                for gov in governors:
                    if gov != pref_gov:
                        gov_cpus = to_remove.get(gov, set())
                        to_remove[gov] = gov_cpus
                        gov_cpus.add(cpu)

        for gov, cpus_to_remove in to_remove.items():
            for cpu in cpus_to_remove:
                governor_cpus[gov].remove(cpu)

    def _cpus_to_str(self, cpus: Set[int]):
        return ','.join((str(c)for c in cpus))

    async def run(self, context: PostProcessContext):
        async with self._cpufreq_man.lock():
            governor_cpus, cpu_governors = self.map_governors(context.restorable_cpus)

            if governor_cpus:
                if len(governor_cpus) == 1:
                    governor = [*governor_cpus][0]
                    cpus = governor_cpus[governor]
                    self._log.debug(f"Restoring CPUs ({self._cpus_to_str(cpus)}) governors to '{governor}'")
                    await self._cpufreq_man.change_governor(governor, cpus)
                else:
                    self._remove_duplicates(governor_cpus, cpu_governors)

                    for governor, cpus in governor_cpus.items():
                        if cpus:
                            self._log.debug(f"Restoring CPUs ({self._cpus_to_str(cpus)}) governors to '{governor}'")
                            await self._cpufreq_man.change_governor(governor, cpus)

            else:
                self._log.warning('Previous CPU governors could be restored because they are unknown')


class PostStopProcesses(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._log = context.logger

    def should_run(self, context: PostProcessContext) -> bool:
        return bool(context.pids_to_stop)

    async def run(self, context: PostProcessContext):
        self._log.debug("Finding children of related processes")
        children = await system.find_children({*context.pids_to_stop})

        if children:
            self._log.debug(f"Children of related processes found: {' '.join([str(p) for p in children])}")
        else:
            children = []
            self._log.debug("No children of related processes found")

        all_to_stop = ' '.join((str(p)for p in (*children, *context.pids_to_stop)))
        self._log.info(f'Stopping related processes: {all_to_stop}')

        code, _ = await system.async_syscall(f'kill -9 {all_to_stop}', return_output=False)

        if code != 0:
            self._log.error(f'Not all related processes could be stopped: {all_to_stop}')


class ReEnableWindowCompositor(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._log = context.logger
        self._context = context

    def should_run(self, context: PostProcessContext) -> bool:
        return bool(context.restore_compositor and self._context.compositor and self._context.compositor_disabled_context is not None)

    async def run(self, context: PostProcessContext):
        compositor, compositor_context = self._context.compositor, self._context.compositor_disabled_context

        async with compositor.lock():
            enabled = await compositor.is_enabled(user_id=context.user_id, user_env=context.user_env, context=compositor_context)

            if enabled is None:
                self._log.error("Could not re-enable the window compositor. It was not possible to determine its current state")
                return
            elif enabled:
                self._log.info("It was not necessary to enable the window compositor. It is already enabled.")
                self._context.compositor_disabled_context = None  # resetting the global context
            else:
                if await compositor.enable(user_id=context.user_id, user_env=context.user_env, context=compositor_context):
                    self._log.info("Window compositor re-enabled")
                    self._context.compositor_disabled_context = None  # resetting the global context
                else:
                    self._log.error("Could not re-enable the window compositor")


class RunFinishScripts(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._context = context
        self._log = context.logger
        self._task = RunScripts('finish', context.allow_root_scripts, self._log)

    def should_run(self, context: PostProcessContext) -> bool:
        if context.scripts:
            for settings in context.scripts:
                if settings.scripts:
                    return True

        return False

    async def run(self, context: PostProcessContext):
        await self._task.run(scripts=context.scripts, user_id=context.user_id, user_env=context.user_env)


class RelaunchStoppedProcesses(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._context = context
        self._log = context.logger
        self._re_python_cmd: Optional[Pattern] = None

    def get_python_cmd_pattern(self) -> Pattern:
        if self._re_python_cmd is None:
            self._re_python_cmd = re.compile(r'^/.+/python\d*\s+(/.+)$')

        return self._re_python_cmd

    def should_run(self, context: PostProcessContext) -> bool:
        return bool(context.stopped_processes and context.user_id is not None)

    async def _run_command(self, name: str, cmd: str):
        try:
            await system.async_syscall(cmd, return_output=False, wait=False)
            self._log.info(f"Process '{name}' ({cmd}) relaunched")
        except:
            stack_log = traceback.format_exc().replace('\n', ' ')
            self._log.warning(f"An exception happened when relaunching process '{name}' ({cmd}): {stack_log}")

    def _run_user_command(self, name: str, cmd: str, user_id: int, user_env: Optional[Dict[str, str]] = None):
        try:
            Process(daemon=True, target=system.run_user_command, kwargs={'cmd': cmd, 'user_id': user_id, 'env': user_env, 'wait': False}).start()
            self._log.info(f"Process '{name}' ({cmd}) relaunched (user={user_id})")
        except:
            stack_log = traceback.format_exc().replace('\n', ' ')
            self._log.warning(f"An exception happened when relaunching process '{name}' ({cmd}) [user={user_id}]: {stack_log}")

    async def run(self, context: PostProcessContext):
        self_is_root = is_root_user()
        root_request = is_root_user(context.user_id)

        if not self_is_root and root_request:
            self._log.warning(f"It will not be possible to launch the following root processes: {', '.join((c[0] for c in context.stopped_processes))}")
            return

        running_cmds = await system.find_processes_by_command({p[1] for p in context.stopped_processes})

        for comm_cmd in context.stopped_processes:
            name, cmd = comm_cmd[0], comm_cmd[1]
            if running_cmds and cmd in running_cmds:
                self._log.warning(f"Process '{name}' ({cmd}) is alive. Skipping its relaunching.")
                continue

            python_cmd = self.get_python_cmd_pattern().findall(cmd)

            real_cmd = python_cmd[0] if python_cmd else cmd

            if self_is_root:
                if root_request:
                    await self._run_command(name, real_cmd)
                else:
                    self._run_user_command(name, real_cmd, context.user_id, context.user_env)
            else:
                await self._run_command(name, real_cmd)


class RestoreMouseCursor(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._log = context.logger
        self._mouse_man = context.mouse_man

    def should_run(self, context: PostProcessContext) -> bool:
        return bool(context.restore_mouse_cursor)

    async def run(self, context: PostProcessContext):
        await self._mouse_man.show_cursor()


class RestoreCPUEnergyPolicyLevel(PostProcessTask):

    def __init__(self, context: OptimizationContext):
        self._log = context.logger
        self._man = context.cpuenergy_man

    def should_run(self, context: PostProcessContext):
        return context.restore_cpu_energy_policy

    async def run(self, context: PostProcessContext):
        async with self._man.lock():
            saved_state = self._man.saved_state

            if not saved_state:
                self._log.info("No CPU energy policy level saved state to restore")
                return

            self._log.info(f"Restoring CPUs energy policy levels: "
                           f"{', '.join(f'{idx}={state}' for idx, state in sorted(saved_state.items()))}")

            cpus_changed = await self._man.change_states(saved_state)

            if not cpus_changed:
                self._log.error("Could not restore CPUs energy policy levels")
                return

            restored, not_restored = [], []

            for idx, changed in cpus_changed.items():
                if changed:
                    restored.append(idx)
                else:
                    not_restored.append(str(idx))

            if not_restored:
                self._log.warning(f"Could not restore the energy policy levels of the following CPUs: "
                                  f"{', '.join(sorted(not_restored))}")

            if restored:
                self._man.clear_state(*restored)
                self._log.debug(f"Saved CPUs energy policy levels cleared: "
                                f"{', '.join(str(i) for i in sorted(restored))}")


class PostProcessTaskManager:

    __ORDER: Dict[Type[PostProcessTask], int] = {ReEnableWindowCompositor: 0,
                                                 PostStopProcesses: 1,
                                                 RestoreMouseCursor: 2,
                                                 RestoreGPUState: 3,
                                                 RestoreCPUGovernor: 4,
                                                 RestoreCPUEnergyPolicyLevel: 5,
                                                 RelaunchStoppedProcesses: 6,
                                                 RunFinishScripts: 7}

    def __init__(self, context: OptimizationContext, tasks: Optional[List[PostProcessTask]] = None):
        self._tasks = tasks if tasks else [cls(context) for cls in PostProcessTask.__subclasses__() if cls != self.__class__]
        self._tasks.sort(key=self._sort)

    def _sort(self, task: PostProcessTask) -> int:
        return self.__ORDER.get(task.__class__, 100)

    def get_available_tasks(self):
        return [*self._tasks]

    def create_tasks(self, context: PostProcessContext) -> Optional[List[Awaitable]]:
        if self._tasks:
            to_run = [t.run(context) for t in self._tasks if t.should_run(context)]
            if to_run:
                return to_run



from abc import ABC, abstractmethod
from typing import List, Optional, Dict, Set, Type, Tuple

from guapow.common.model import ScriptSettings
from guapow.service.optimizer.gpu import GPUState, GPUDriver
from guapow.service.optimizer.task.model import CPUState, OptimizedProcess, OptimizationContext


class PostProcessSummary:

    def __init__(self, pids_alive: Optional[Set[int]], user_id: Optional[int], user_env: Optional[Dict[str, str]], restore_compositor: Optional[bool],
                 previous_cpu_states: Optional[List[CPUState]], previous_gpus_states: Optional[Dict[Type[GPUDriver], List[GPUState]]],
                 pids_to_stop: Optional[Set[int]], processes_relaunch_by_time: Optional[Dict[float, Dict[str, str]]],
                 post_scripts: Optional[Dict[float, ScriptSettings]], keep_compositor_disabled: Optional[bool], cpus_in_use: Optional[bool],
                 gpus_in_use: Optional[Dict[Type[GPUDriver], Set[int]]], processes_not_relaunch: Optional[Set[str]],
                 processes_to_relaunch: Optional[Dict[str, str]], dead_pids: Optional[Set[Tuple[int, int]]], restore_mouse_cursor: Optional[bool],
                 keep_mouse_hidden: Optional[bool], keep_cpu_energy_policy: Optional[bool],
                 restore_cpu_energy_policy: Optional[bool]):
        self.pids_alive = pids_alive
        self.user_id = user_id
        self.user_env = user_env
        self.previous_cpus_states = previous_cpu_states
        self.previous_gpus_states = previous_gpus_states
        self.pids_to_stop = pids_to_stop
        self.processes_relaunch_by_time = processes_relaunch_by_time
        self.post_scripts = post_scripts
        self.keep_compositor_disabled = keep_compositor_disabled
        self.restore_compositor = restore_compositor
        self.cpus_in_use = cpus_in_use
        self.gpus_in_use = gpus_in_use
        self.processes_not_relaunch = processes_not_relaunch
        self.processes_to_relaunch = processes_to_relaunch
        self.dead_pids = dead_pids
        self.restore_mouse_cursor = restore_mouse_cursor
        self.keep_mouse_hidden = keep_mouse_hidden
        self.keep_cpu_energy_policy = keep_cpu_energy_policy
        self.restore_cpu_energy_policy = restore_cpu_energy_policy

    @classmethod
    def empyt(cls) -> "PostProcessSummary":
        return cls(pids_alive=None, user_id=None, user_env=None, restore_compositor=None, previous_cpu_states=None,
                   cpus_in_use=None, previous_gpus_states=None, pids_to_stop=None, processes_relaunch_by_time=None,
                   post_scripts=None, gpus_in_use=None, keep_compositor_disabled=None, processes_not_relaunch=None,
                   processes_to_relaunch=None, dead_pids=None, keep_mouse_hidden=None, restore_mouse_cursor=None,
                   keep_cpu_energy_policy=None, restore_cpu_energy_policy=None)


class PostProcessSummarizer(ABC):

    @abstractmethod
    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        pass


class UserEnvironmentSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if not process.alive and process.user_env:
            if summary.user_env is None:
                summary.user_env = {}

            summary.user_env.update(process.user_env)


class UserIdSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if not process.alive and process.user_id is not None:
            summary.user_id = process.user_id


class ProcessesToStopSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if not process.alive and process.related_pids and summary.pids_alive:
            if summary.pids_to_stop is None:
                summary.pids_to_stop = set()

            summary.pids_to_stop.update({pid for pid in process.related_pids if pid in summary.pids_alive})


class MouseCursorStateSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if process.requires_mouse_hidden:
            if process.alive:
                summary.keep_mouse_hidden = True
            elif await main_context.is_mouse_cursor_hidden():  # if the mouse cursor was hidden by the Optimizer
                summary.restore_mouse_cursor = True


class CompositorStateSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if process.requires_compositor_disabled:
            if process.alive:
                summary.keep_compositor_disabled = True
            elif main_context.compositor_disabled_context is not None:
                summary.restore_compositor = True


class CPUGovernorStateSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if process.previous_cpu_state:
            if process.alive:
                summary.cpus_in_use = True
            else:
                if summary.previous_cpus_states is None:
                    summary.previous_cpus_states = []

                summary.previous_cpus_states.append(process.previous_cpu_state)


class CPUEnergyPolicyLevelSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if process.cpu_energy_policy_changed:
            if process.alive:
                summary.keep_cpu_energy_policy = True
            else:
                summary.restore_cpu_energy_policy = True


class GPUStateSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if process.previous_gpus_states:
            if process.alive:
                if summary.gpus_in_use is None:
                    summary.gpus_in_use = {}

                for driver, states in process.previous_gpus_states.items():
                    ids = summary.gpus_in_use.get(driver, set())
                    ids.update({s.id for s in states})
                    summary.gpus_in_use[driver] = ids
            else:
                if summary.previous_gpus_states is None:
                    summary.previous_gpus_states = {}

                for driver, gpus_states in process.previous_gpus_states.items():
                    states = summary.previous_gpus_states.get(driver, [])
                    summary.previous_gpus_states[driver] = states
                    states.extend(gpus_states)


class ProcessesToRelaunchSummarizer(PostProcessSummarizer):

    def _fill(self, summary: PostProcessSummary, process: OptimizedProcess, processes: Dict[str, Optional[str]], relaunch: Optional[bool]):
        if process.alive:
            if summary.processes_not_relaunch is None:
                summary.processes_not_relaunch = set()

            summary.processes_not_relaunch.update(processes)
        else:
            if summary.processes_relaunch_by_time is None:
                summary.processes_relaunch_by_time = {}

            if relaunch:
                if summary.processes_to_relaunch:  # to always get the appropriate command (the clients might not have been able to inform the command, because the process was dead)
                    summary.processes_relaunch_by_time[process.created_at] = {comm: summary.processes_to_relaunch.get(comm, cmd) for comm, cmd in processes.items()}
                else:
                    summary.processes_relaunch_by_time[process.created_at] = processes
            elif summary.processes_to_relaunch:
                summary.processes_relaunch_by_time[process.created_at] = {comm: cmd for comm, cmd in processes.items() if
                                                                          comm in summary.processes_to_relaunch}

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if process.stopped_processes:
            self._fill(summary, process, process.stopped_processes, process.relaunch_stopped_processes)

        if process.stopped_after_launch:
            self._fill(summary, process, process.stopped_after_launch, process.relaunch_stopped_after_launch)


class FinishScriptsSummarizer(PostProcessSummarizer):

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if not process.alive and process.post_scripts:
            if summary.post_scripts is None:
                summary.post_scripts = {}

            summary.post_scripts[process.created_at] = process.post_scripts


class GeneralPostProcessSummarizer(PostProcessSummarizer):

    ANALYSER_ORDER: Dict[Type[PostProcessSummarizer], int] = {
        UserIdSummarizer: 0,
        UserEnvironmentSummarizer: 1,
        ProcessesToStopSummarizer: 2,
        CompositorStateSummarizer: 3,
        FinishScriptsSummarizer: 4,
        CPUGovernorStateSummarizer: 5,
        CPUEnergyPolicyLevelSummarizer: 6,
        GPUStateSummarizer: 7,
        ProcessesToRelaunchSummarizer: 8
    }

    __instance = None

    def __init__(self, fillers: Optional[List[PostProcessSummarizer]]):
        self._fillers = fillers

    @classmethod
    def sort_filler(cls, filler: PostProcessSummarizer) -> int:
        return cls.ANALYSER_ORDER.get(filler.__class__, 99)

    def get_fillers(self) -> Optional[List[PostProcessSummarizer]]:
        return [*self._fillers] if self._fillers else None

    @classmethod
    def instance(cls) -> "GeneralPostProcessSummarizer":
        if cls.__instance is None:
            fillers = [sub() for sub in PostProcessSummarizer.__subclasses__() if sub != cls]
            fillers.sort(key=cls.sort_filler)
            cls.__instance = cls(fillers)

        return cls.__instance

    async def fill(self, summary: PostProcessSummary, process: OptimizedProcess, main_context: OptimizationContext):
        if self._fillers:
            for analyser in self._fillers:
                await analyser.fill(summary, process, main_context)

    async def summarize(self, processes: List[OptimizedProcess], pids_alive: Set[int], processes_to_relaunch: Optional[Dict[str, Optional[str]]], context: OptimizationContext) -> PostProcessSummary:
        summary = PostProcessSummary.empyt()
        summary.processes_to_relaunch = processes_to_relaunch
        summary.pids_alive = pids_alive

        if processes:
            for idx, process in enumerate(processes):
                if not pids_alive or process.pid not in summary.pids_alive:
                    process.alive = False

                    if summary.dead_pids is None:
                        summary.dead_pids = set()

                    summary.dead_pids.add((idx, process.pid))

                await self.fill(summary, process, context)

        return summary



import functools
from abc import ABC, abstractmethod
from logging import Logger
from typing import List, Optional, Set, Tuple, Type, Dict

from guapow.common import class_util
from guapow.common.model import CustomEnum, ProfileFile, FileModel


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split('.'))


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


class InvalidMappedPropertyException(Exception):

    def __init__(self, msg: str):
        self.message = msg


class FileModelPropertyMapper(ABC):

    @abstractmethod
    def supports(self, prop_type: type) -> bool:
        pass

    @abstractmethod
    def map(self, prop_val: str, prop_type: type) -> Optional[object]:
        pass

    @abstractmethod
    def get_raw_type(self) -> type:
        pass


class FileModelCollectionPropertyMapper(FileModelPropertyMapper, ABC):

    @abstractmethod
    def create_collection(self) -> object:
        pass

    @abstractmethod
    def update_collection(self, collection: object, update: object):
        pass


class StringPropertyMapper(FileModelPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == str

    def map(self, prop_val: str, prop_type: type) -> Optional[str]:
        return prop_val

    def get_raw_type(self) -> type:
        return str


class IntPropertyMapper(FileModelPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == int

    def map(self, prop_val: str, prop_type: type) -> Optional[int]:
        try:
            return int(prop_val)
        except ValueError:
            raise InvalidMappedPropertyException("It should be an integer")

    def get_raw_type(self) -> type:
        return int


class FloatPropertyMapper(FileModelPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == float

    def map(self, prop_val: str, prop_type: type) -> Optional[float]:
        try:
            return float(prop_val)
        except ValueError:
            raise InvalidMappedPropertyException("It should be a float")

    def get_raw_type(self) -> type:
        return float


class BoolPropertyMapper(FileModelPropertyMapper):

    FALSE_STRINGS = {'0', 'false'}
    TRUE_STRINGS = {'1', 'true'}

    def supports(self, prop_type: type) -> bool:
        return prop_type == bool

    def map(self, prop_val: str, prop_type: type) -> Optional[bool]:
        lower_val = prop_val.lower()

        if lower_val in self.FALSE_STRINGS:
            return False
        elif lower_val in self.TRUE_STRINGS:
            return True
        else:
            raise InvalidMappedPropertyException(f"It should be a boolean (accepted values: {','.join([*self.FALSE_STRINGS, *self.TRUE_STRINGS])})")

    def get_raw_type(self) -> type:
        return bool


class IntListPropertyMapper(FileModelCollectionPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == List[int]

    def map(self, prop_val: str, prop_type: type) -> Optional[List[int]]:
        int_list = [*{int(n) for n in prop_val.split(',') if n.isdigit()}]
        int_list.sort()
        return int_list

    def create_collection(self) -> object:
        return list()

    def update_collection(self, collection: object, update: object):
        if isinstance(collection, list) and isinstance(update, list):
            collection.extend(update)

    def get_raw_type(self) -> type:
        return list


class IntSetPropertyMapper(FileModelCollectionPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == Set[int]

    def map(self, prop_val: str, prop_type: type) -> Optional[Set[int]]:
        return {int(n) for n in prop_val.split(',') if n.isdigit()}

    def create_collection(self) -> object:
        return set()

    def update_collection(self, collection: object, update: object):
        if isinstance(collection, set) and isinstance(update, set):
            collection.update(update)

    def get_raw_type(self) -> type:
        return set


class StringListPropertyMapper(FileModelCollectionPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == List[str]

    def map(self, prop_val: str, prop_type: type) -> Optional[List[str]]:
        return prop_val.split(',')

    def create_collection(self) -> object:
        return list()

    def update_collection(self, collection: object, update: object):
        if isinstance(collection, list) and isinstance(update, list):
            collection.extend(update)

    def get_raw_type(self) -> type:
        return list


class StringSetPropertyMapper(FileModelCollectionPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == Set[str]

    def map(self, prop_val: str, prop_type: type) -> Optional[Set[str]]:
        return {*prop_val.split(',')}

    def create_collection(self) -> object:
        return set()

    def update_collection(self, collection: object, update: object):
        if isinstance(collection, set) and isinstance(update, set):
            collection.update(update)

    def get_raw_type(self) -> type:
        return set


class CustomEnumPropertyMapper(FileModelPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return isinstance(prop_type, type) and issubclass(prop_type, CustomEnum)

    def map(self, prop_val: str, prop_type: Type[CustomEnum]) -> Optional[CustomEnum]:
        mapped_val = prop_type.from_str(prop_val)

        if mapped_val:
            return mapped_val

        raise InvalidMappedPropertyException(f'Unknown value: {prop_val}')

    def get_raw_type(self) -> type:
        return CustomEnum


class DictPropertyMapper(FileModelCollectionPropertyMapper):

    def supports(self, prop_type: type) -> bool:
        return prop_type == dict

    def map(self, prop_val: str, prop_type: type) -> Optional[Tuple[str, object]]:
        var_val = [s.strip() for s in prop_val.split(':', 1)]

        if not var_val or not var_val[0] or len(var_val) > 2:
            raise InvalidMappedPropertyException(f'Wrong dict property value defined: {prop_val}')

        if len(var_val) == 1 and var_val[0]:
            return var_val[0], None
        elif len(var_val) == 2 and not var_val[1]:
            return var_val[0], None
        else:
            return var_val[0], var_val[1]

    def create_collection(self) -> object:
        return dict()

    def update_collection(self, collection: object, update: object):
        if isinstance(collection, dict):
            if isinstance(update, tuple):
                collection[update[0]] = update[1]
            elif isinstance(update, dict):
                collection.update(update)

    def get_raw_type(self) -> type:
        return dict


class FileModelFiller:

    def __init__(self, logger: Logger):
        self._log = logger
        self._property_mappers = class_util.instantiate_subclasses(FileModelPropertyMapper)
        self._mapper_type_cache: Dict[type, FileModelPropertyMapper] = {}

    def get_mapper(self, prop_type: type) -> Optional[FileModelPropertyMapper]:
        mapper = self._mapper_type_cache.get(prop_type)

        if not mapper:
            supported_mappers = [m for m in self._property_mappers if m.supports(prop_type)]

            if supported_mappers:
                mapper = supported_mappers[0]
                self._mapper_type_cache[prop_type] = mapper

        return mapper

    def _update_mapped_object_with_mapper(self, prop_path: str, prop_val: str, prop_type: type, root: FileModel, mapper: FileModelPropertyMapper, mapped_obj: dict):
        try:
            mapped_val = mapper.map(prop_val, prop_type)
        except InvalidMappedPropertyException as e:
            self._log.warning(f"Invalid {root.get_output_name()}'s property '{prop_path}' mapping: {e.message}")
            return

        if isinstance(mapper, FileModelCollectionPropertyMapper):
            if mapped_val:
                collection = mapped_obj.get(prop_path)

                if collection is None:
                    collection = mapper.create_collection()
                    mapped_obj[prop_path] = collection

                mapper.update_collection(collection, mapped_val)
        else:
            mapped_obj[prop_path] = mapped_val

    def fill(self, root: FileModel, file_content: str, only_properties: Optional[Set[str]] = None):
        prop_mapping = root.get_full_mapping()

        if prop_mapping:
            mapped_obj = {}

            for line in file_content.split('\n'):
                if line:
                    clean_line = line.strip()

                    if clean_line and not clean_line.startswith('#'):
                        line_split = clean_line.split('=', 1)
                        file_prop = line_split[0].strip()

                        if file_prop:
                            if not only_properties or file_prop in only_properties:
                                model_prop = prop_mapping.get(file_prop)

                                if model_prop:
                                    prop_path, prop_type = model_prop[0], model_prop[1]

                                    if len(line_split) == 1:
                                        prop_default = model_prop[2]

                                        if prop_default is None:
                                            continue

                                        mapper = self.get_mapper(prop_type)

                                        if not mapper or not isinstance(prop_default, mapper.get_raw_type()):
                                            self._log.error(f"Invalid default value type for property '{file_prop}' ({prop_type.__name__}): {prop_default} ({type(prop_default).__name__})")
                                            continue

                                        mapped_obj[prop_path] = prop_default
                                    else:
                                        mapper = self.get_mapper(prop_type)

                                        if mapper:
                                            prop_val = line_split[1].split('#')[0].strip()
                                            self._update_mapped_object_with_mapper(prop_path, prop_val, prop_type, root, mapper, mapped_obj)
                                        else:
                                            self._log.error(f"Unsupported {root.get_output_name()}'s property type: {prop_type}")

            if mapped_obj:
                for prop, val in mapped_obj.items():
                    rsetattr(root, prop, val)

    def fill_profile(self, profile: ProfileFile, profile_str: str, profile_path: Optional[str], add_settings: Optional[str] = None):
        profile.set_path(profile_path)
        final_profile_str = profile_str

        if add_settings:
            final_profile_str = f'{final_profile_str}\n{add_settings}'
            self._log.debug(f"Settings merged into profile '{profile.name}': {add_settings}")

        self.fill(profile, final_profile_str)
        profile.reset_invalid_nested_members()



import re
from re import Pattern
from typing import Optional

re_any_operator = re.compile(r'\*+')


def map_any_regex(word: str) -> Optional[Pattern]:
    if word:
        if has_any_regex(word):
            return map_only_any_regex(word)
        else:
            return re.compile(r'^{}$'.format(re.escape(word)))


def has_any_regex(word: str) -> bool:
    return '*' in word if word else False


def map_only_any_regex(word: str) -> Optional[Pattern]:
    escaped_word = re.escape(re_any_operator.sub('@', word))
    return re.compile(r'^{}$'.format(escaped_word.replace('@', '.+')))


def strip_file_extension(cmd: str) -> Optional[str]:
    if cmd:
        return '.'.join(cmd.split('.')[0:-1])




import asyncio
import os
import shutil
from abc import ABC
from io import StringIO
from typing import Optional, Tuple

from guapow.common.system import async_syscall
from guapow.common.users import is_root_user
from guapow.service.optimizer.renicer import Renicer
from guapow.service.optimizer.task.model import Task, OptimizationContext, OptimizedProcess


class ProcessTask(Task, ABC):
    """
    Task that requires a process to be executed
    """


class ReniceProcess(ProcessTask):

    def __init__(self, context: OptimizationContext):
        super(ReniceProcess, self).__init__(context)
        self._log = context.logger
        self._renicer = Renicer(logger=context.logger, watch_interval=context.renicer_interval)

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    async def should_run(self, process: OptimizedProcess) -> bool:
        request, profile = process.request, process.profile

        if profile.process and profile.process.nice and profile.process.nice.level is not None:
            if profile.process.nice.has_valid_level():
                return True
            else:
                self._log.warning(f"Invalid nice level ({profile.process.nice.level}) defined for {profile.get_log_str()}. "
                                  f"Valid values between -20 and 20. Process ({process.pid}) will not be reniced "
                                  f"(request={process.request.pid})")

        return False

    async def run(self, process: OptimizedProcess):
        profile = process.profile

        if profile.process.nice.delay is not None:
            if profile.process.nice.delay > 0:
                self._log.info(f"Delaying process '{process.pid}' renicing for {profile.process.nice.delay} seconds "
                               f"(request={process.request.pid})")
                await asyncio.sleep(profile.process.nice.delay)
            else:
                self._log.warning(f"Invalid nice delay defined for process '{process.pid}': {profile.process.nice.delay} "
                                  f"(must be higher than zero) (request={process.request.pid})")

        self._renicer.set_priority(process.pid, profile.process.nice.level, process.request.pid)

        if process.profile.process.nice.watch:
            if self._renicer.add(process.pid, profile.process.nice.level, process.request.pid):
                self._renicer.watch()


class ChangeCPUAffinity(ProcessTask):

    def __init__(self, context: OptimizationContext):
        super(ChangeCPUAffinity, self).__init__(context)
        self._log = context.logger
        self._cpu_count = context.cpu_count

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        if self._cpu_count == 0:
            return False, "no CPUs detected. It will not be possible to change CPU affinity"

        return True, None

    async def should_run(self, process: OptimizedProcess) -> bool:
        profile = process.profile

        if profile.process and profile.process.cpu_affinity:
            if profile.process.has_valid_cpu_affinity(self._cpu_count):
                return True
            else:
                self._log.warning(f"Invalid CPU affinity defined ({profile.process.cpu_affinity}) for {profile.get_log_str()}. "
                                  f"It must be a list of integers between '0' and '{self._cpu_count - 1}' (request={process.request.pid})")

        return False

    async def run(self, process: OptimizedProcess):
        profile = process.profile

        try:
            os.sched_setaffinity(process.pid, profile.process.cpu_affinity)
            self._log.info(f"Process '{process.pid}' CPU affinity changed to {profile.process.cpu_affinity} (request={process.request.pid})")
        except:
            self._log.error(f"Could not change process '{process.pid}' CPU affinity to {profile.process.cpu_affinity} (request={process.request.pid})")


class ChangeCPUScalingPolicy(ProcessTask):

    def __init__(self, context: OptimizationContext):
        super(ChangeCPUScalingPolicy, self).__init__(context)
        self._log = context.logger
        self._cpu_count = context.cpu_count

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        return True, None

    async def should_run(self, process: OptimizedProcess) -> bool:
        request, profile = process.request, process.profile

        if profile.process and profile.process.scheduling and profile.process.scheduling.policy:
            sched = profile.process.scheduling

            if sched.policy.requires_root() and not is_root_user():
                self._log.warning(f"Not possible to change the scheduling policy for process '{process.pid}' "
                                  f"to '{profile.process.scheduling.policy.name}'. It requires root privileges. (request={process.request.pid})")
                return False
            elif sched.policy.requires_priority() and sched.priority is not None and not sched.has_valid_priority():
                self._log.warning(f"Invalid priority '{sched.priority}' defined for scheduling policy '{sched.policy.name}' "
                                  f"in {profile.get_log_str()} (request={process.request.pid})")
                return False

            return True

        return False

    async def run(self, process: OptimizedProcess):
        profile, scheduling = process.profile, process.profile.process.scheduling

        priority = 0
        if scheduling.policy.requires_priority():
            priority = scheduling.priority

            if scheduling.priority is None:
                priority = 1
                self._log.warning(f"No priority set for policy '{scheduling.policy.name}' in {profile.get_log_str()}. "
                                  f"('{priority}' will be used) (request={process.request.pid})")

        elif scheduling.priority is not None:
            self._log.warning(f"Scheduling policy '{scheduling.policy.name}' does not require priority "
                              f"('{scheduling.priority}' will be ignored) (request={process.request.pid})")

        try:
            os.sched_setscheduler(process.pid, scheduling.policy.value(), os.sched_param(priority))
            self._log.info(f"Process '{process.pid}' scheduling policy changed to '{scheduling.policy.name}'"
                           f"{f' (priority: {priority})' if priority else ''} (request={process.request.pid})")
        except:
            self._log.error(f"Could not change process '{process.pid}' scheduling policy to '{scheduling.policy.name}'"
                            f"{f' (priority: {priority})' if priority else ''} (request={process.request.pid})")


class ChangeProcessIOClass(ProcessTask):

    def __init__(self, context: OptimizationContext):
        super(ChangeProcessIOClass, self).__init__(context)
        self._log = context.logger

    async def is_available(self) -> Tuple[bool, Optional[str]]:
        if shutil.which('ionice'):
            return True, None
        else:
            return False, "'ionice' is not installed. It will not be possible to change a process IO scheduling"

    async def should_run(self, process: OptimizedProcess) -> bool:
        profile = process.profile

        if profile.process.io and profile.process.io.ioclass:
            if profile.process.io.ioclass.supports_priority():

                # if nice_level is None, 0 will be considered
                if profile.process.io.nice_level is None or profile.process.io.has_valid_priority():
                    return True
                else:
                    self._log.warning(f"Invalid IO nice level ({profile.process.io.nice_level}) defined for {profile.get_log_str()}. "
                                      f"It must be a value between 0 and 7. IO class will not be changed (request={process.request.pid})")
                    return False

            return True

        return False

    async def run(self, process: OptimizedProcess):
        io_config = process.profile.process.io

        if io_config.ioclass.supports_priority():
            if io_config.nice_level is None:
                self._log.warning(f"No nice level defined for IO class '{io_config.ioclass.name}' on {process.profile.get_log_str()}. "
                                  f"'0' will be considered (request={process.request.pid})")
                priority = 0
            else:
                priority = io_config.nice_level
        else:
            priority = None

        cmd = StringIO()
        cmd.write(f'ionice -p {process.pid} -c {io_config.ioclass.value()}')

        if priority is not None:
            cmd.write(f' -n {priority}')

        cmd.seek(0)
        cmd_str = cmd.read()

        self._log.info(f"Changing process '{process.pid}' IO class to '{io_config.ioclass.name}'"
                       f"{f' (priority: {priority})'if priority is not None else ''} (request={process.request.pid}): {cmd_str}")

        exitcode, output = await async_syscall(cmd_str)

        if exitcode != 0:
            self._log.error(f"Could not change process '{process.pid}' IO class for {process.profile.get_log_str()} "
                            f"(request={process.request.pid})")

            if output:
                for line in output.split('\n'):
                    self._log.error(line)



import asyncio
import getpass
import os
import sys
import time
from subprocess import Popen

from guapow import __app_name__
from guapow.common import network
from guapow.common.auth import read_machine_id
from guapow.common.config import read_optimizer_config
from guapow.common.dto import OptimizationRequest
from guapow.common.log import new_logger
from guapow.common.model_util import FileModelFiller
from guapow.runner.profile import RunnerProfileReader
from guapow.runner.settings import is_log_enabled, read_valid_full_config, read_additional_profile_config, \
    InvalidConfigurationException, is_file_log
from guapow.runner.task import RunnerTaskManager, RunnerContext


def run():
    asyncio.get_event_loop().run_until_complete(launch_process())


async def launch_process():
    lti = time.time()
    if len(sys.argv) < 2:
        sys.stderr.write('Command not informed. Aborting...\n')
        exit(1)
    else:
        user_id, user_name, profile = os.getuid(), getpass.getuser(), os.getenv('GUAPOW_PROFILE')
        log_to_file = is_file_log()
        logger = new_logger(name=__app_name__, service=False, enabled=log_to_file or is_log_enabled(),
                            write_to_file=log_to_file, threaded=False)

        try:
            config_str, profile_config_str = read_valid_full_config(logger), None
        except InvalidConfigurationException:
            return exit(1)

        model_filler = FileModelFiller(logger)
        profile_reader = RunnerProfileReader(model_filler, logger)

        if config_str:
            runner_profile = profile_reader.map_valid_config(config=config_str)
        else:
            profile_config_str = read_additional_profile_config(logger)
            runner_profile = await profile_reader.read_available(user_id=user_id, user_name=user_name, profile=profile, add_settings=profile_config_str)

        context, extra_proc_args = None, {}

        if runner_profile:
            context = RunnerContext(logger=logger, processes_initialized=set(), environment_variables={}, stopped_processes={})
            await RunnerTaskManager(context).run(runner_profile)
            extra_proc_args.update(context.get_process_extra_arguments())

        cmd_str = ' '.join(sys.argv[1:])
        logger.info(f'Launching command: {cmd_str}')

        try:
            proc = Popen(sys.argv[1:], **extra_proc_args)
        except OSError as e:
            sys.stderr.write(f"An error occurred when launching: {' '.join(sys.argv[1:])}\n")
            sys.stderr.write(str(e))
            return exit(1)

        ltf = time.time()
        logger.info(f'Launch time: {ltf - lti:.4f} seconds')

        common_attrs = {'pid': proc.pid, 'command': cmd_str, 'user_name': user_name, 'created_at': time.time(),
                        'user_env': extra_proc_args['env'] if extra_proc_args and extra_proc_args['env'] else dict(os.environ)}

        if context:
            if context.processes_initialized:
                common_attrs['related_pids'] = context.processes_initialized

            if context.stopped_processes:
                common_attrs['stopped_processes'] = context.stopped_processes

                if runner_profile.stop.relaunch:
                    common_attrs['relaunch_stopped_processes'] = True

        if config_str:
            request = OptimizationRequest.new_from_config(config=config_str, **common_attrs)
        else:
            request = OptimizationRequest.new_from_profile(profile=profile, profile_config=profile_config_str, **common_attrs)

        opt_config = await read_optimizer_config(user_id=user_id, user_name=user_name, logger=logger,
                                                 filler=model_filler, only_properties={'port', 'request.encrypted'})

        if not opt_config:
            logger.warning("Invalid Optimizer config. It will not be possible to send the request")
            exit(1)

        machine_id = await read_machine_id() if opt_config.encrypted_requests else None

        if not machine_id:
            logger.warning("Encryption is disabled")

        await network.send(request=request, opt_config=opt_config, machine_id=machine_id, logger=logger)


if __name__ == '__main__':
    run()



import asyncio
import re
import time
from asyncio import create_task
from logging import Logger
from typing import Dict, Optional, Tuple, Set, Any, Union

from guapow.common import network
from guapow.common.config import OptimizerConfig
from guapow.common.dto import OptimizationRequest
from guapow.service.watcher import mapping, ignored
from guapow.service.watcher.config import ProcessWatcherConfig
from guapow.service.watcher.patterns import RegexMapper, RegexType
from guapow.service.watcher.util import map_processes


class ProcessWatcherContext:

    def __init__(self, user_id: int, user_name: str, user_env: Dict[str, str], logger: Logger,
                 optimized: Dict[int, str], opt_config: OptimizerConfig, watch_config: ProcessWatcherConfig,
                 mapping_file_path: str, machine_id: Optional[str], ignored_file_path: str,
                 ignored_procs: Dict[Union[str, re.Pattern], Set[str]]):
        self.user_id = user_id
        self.user_name = user_name
        self.user_env = user_env
        self.logger = logger
        self.optimized = optimized  # keeps the state of which processes were optimized to avoid sending duplicate requests
        self.opt_config = opt_config
        self.watch_config = watch_config
        self.mapping_file_path = mapping_file_path
        self.machine_id = machine_id
        self.ignored_file_path = ignored_file_path
        self.ignored_procs = ignored_procs


class ProcessWatcher:

    def __init__(self, regex_mapper: RegexMapper, context: ProcessWatcherContext):
        self._context = context
        self._regex_mapper = regex_mapper
        self._log = context.logger

        # cached attributes
        self._mapping_cached = False
        self._mappings: Optional[Dict[str, str]] = None
        self._cmd_patterns: Optional[Dict[re.Pattern, str]] = None
        self._comm_patterns: Optional[Dict[re.Pattern, str]] = None

        self._last_mapping_file_found: Optional[bool] = None  # controls repetitive file found logs
        self._last_ignored_file_found: Optional[bool] = None  # controls repetitive file found logs

        self._ignored_cached = False
        self._ignored_exact_strs: Optional[Set[str]] = None
        self._ignored_cmd_patterns: Optional[Set[re.Pattern]] = None
        self._ignored_comm_patterns: Optional[Set[re.Pattern]] = None

    async def _read_ignored(self) -> Optional[Tuple[Set[str], Optional[Set[re.Pattern]], Optional[Set[re.Pattern]]]]:
        """
        return a tuple with command patterns (cmd) and name patterns (comm)
        """
        if self._ignored_cached:
            if self._ignored_exact_strs:
                return self._ignored_exact_strs, self._ignored_cmd_patterns, self._ignored_comm_patterns
        else:
            file_found, ignored_strs = await ignored.read(file_path=self._context.ignored_file_path, logger=self._log,
                                                          last_file_found_log=self._last_ignored_file_found)
            self._last_ignored_file_found = file_found

            if not self._ignored_cached and self._context.watch_config.ignored_cache:
                self._log.debug("Caching ignored patterns to memory")
                self._ignored_cached = True  # pre-saving the caching state (if enabled)

            if ignored_strs:
                if self._context.watch_config.ignored_cache:  # caching to memory (if enabled)
                    self._ignored_exact_strs = ignored_strs

                patterns = self._regex_mapper.map_collection(ignored_strs)

                cmd_patterns, comm_patterns = None, None

                if patterns:
                    cmd_patterns, comm_patterns = patterns.get(RegexType.CMD), patterns.get(RegexType.COMM)

                    if self._context.watch_config.ignored_cache:  # caching to memory (if enabled)
                        self._ignored_cmd_patterns = cmd_patterns
                        self._ignored_comm_patterns = comm_patterns

                return ignored_strs, cmd_patterns, comm_patterns

    async def _read_mappings(self) -> Optional[Tuple[Dict[str, str], Optional[Dict[re.Pattern, str]], Optional[Dict[re.Pattern, str]]]]:
        if self._mapping_cached:
            if self._mappings:
                return self._mappings, self._cmd_patterns, self._comm_patterns
        else:
            file_found, mappings = await mapping.read(file_path=self._context.mapping_file_path, logger=self._log, last_file_found_log=self._last_mapping_file_found)
            self._last_mapping_file_found = file_found

            pattern_mappings = self._regex_mapper.map_for_profiles(mappings)
            cmd_patterns, comm_patterns = (pattern_mappings[0], pattern_mappings[1]) if pattern_mappings else (None, None)

            if self._context.watch_config.mapping_cache:
                self._mappings, self._cmd_patterns, self._comm_patterns = mappings, cmd_patterns, comm_patterns
                self._mapping_cached = True

            return mappings, cmd_patterns, comm_patterns

    def _map_ignored_id(self, pid: int, comm: str) -> str:
        return f'{pid}:{comm}'

    def _is_ignored(self, ignored_id: str) -> bool:
        if self._context.ignored_procs:
            for ignored_ids in self._context.ignored_procs.values():
                if ignored_id in ignored_ids:
                    return True

        return False

    def _clean_old_ignore_patterns(self, ignored_exact: Set[str], current_cmd_patterns: Optional[Set[re.Pattern]],
                                   current_comm_patterns: Optional[Set[re.Pattern]]):
        if self._context.ignored_procs:
            all_patterns = (ignored_exact, current_cmd_patterns, current_comm_patterns)
            to_remove = set()
            for pattern in self._context.ignored_procs:
                found = False
                for patterns in all_patterns:
                    if patterns and pattern in patterns:
                        found = True
                        break

                if not found:
                    to_remove.add(pattern)

            if to_remove:
                self._log.debug(f"Cleaning old ignored patterns from context: {', '.join(str(p) for p in to_remove)}")
                for pattern in to_remove:
                    del self._context.ignored_procs[pattern]

    def _matches_ignored(self, cmd_com: Tuple[str, str], ignored_exact: Set[str],
                         cmd_patterns: Optional[Set[re.Pattern]], comm_patterns: Optional[Set[re.Pattern]],
                         ignored_id: str) -> bool:

        if ignored_exact or cmd_patterns or comm_patterns:
            for idx, cmd in enumerate(cmd_com):  # 0: cmd, 1: comm
                matched_pattern = None
                if ignored_exact and cmd in ignored_exact:  # exact matches have higher priority than patterns
                    matched_pattern = cmd
                else:
                    regex_ignored = cmd_patterns if idx == 0 else comm_patterns

                    if regex_ignored:
                        for pattern in regex_ignored:
                            if pattern.match(cmd):
                                matched_pattern = pattern
                                break

                if matched_pattern:
                    ignored_ids = self._context.ignored_procs.get(matched_pattern)

                    if ignored_ids is None:
                        ignored_ids = set()
                        self._context.ignored_procs[matched_pattern] = ignored_ids

                    ignored_ids.add(ignored_id)
                    return True

        return False

    async def check_mappings(self):
        mapping_tuple = await self._read_mappings()

        if not mapping_tuple or not mapping_tuple[0]:
            return

        mappings, cmd_patterns, comm_patterns = mapping_tuple[0], mapping_tuple[1], mapping_tuple[2]

        task_map_procs, task_ignored = create_task(map_processes()), create_task(self._read_ignored())

        procs = await task_map_procs

        if not procs:
            self._log.warning('No processes alive')
            self._context.optimized.clear()
            return

        ignored_exact, ignored_cmd, ignored_comm = None, None, None
        ignored_patterns = await task_ignored

        if ignored_patterns:
            ignored_exact, ignored_cmd, ignored_comm = ignored_patterns

        self._clean_old_ignore_patterns(ignored_exact, ignored_cmd, ignored_comm)

        tasks = []
        for pid, cmd_comm in procs.items():
            ignored_id = self._map_ignored_id(pid, cmd_comm[1])

            if self._is_ignored(ignored_id):
                continue

            previously_optimized_cmd = self._context.optimized.get(pid)

            if previously_optimized_cmd and (previously_optimized_cmd == cmd_comm[0] or previously_optimized_cmd == cmd_comm[1]):
                continue

            if self._matches_ignored(cmd_comm, ignored_exact, ignored_cmd, ignored_comm, ignored_id):
                self._log.info(f"Ignoring process (pid: {pid}, name: {cmd_comm[1]})")
                continue

            for idx, cmd in enumerate(cmd_comm):  # 0: cmd, 1: comm
                profile = mappings.get(cmd)  # exact matches have higher priority than patterns
                if profile:
                    tasks.append(self.send_request(pid, cmd_comm[0], cmd, profile))
                    break  # 'cmd' has higher priority than 'comm'
                else:
                    regex_mapping = cmd_patterns if idx == 0 else comm_patterns

                    if regex_mapping:
                        matches = dict()
                        for pattern, patter_profile in regex_mapping.items():
                            if pattern.match(cmd):
                                matches[pattern.pattern] = patter_profile

                        if matches:
                            match_profile = matches[max(matches.keys())]  # retrieves the more specific match by length
                            tasks.append(self.send_request(pid, cmd_comm[0], cmd, match_profile))
                            break

        if tasks:
            await asyncio.gather(*tasks)

        self._clean_dead_processes_from_context(procs)

    def _clean_dead_processes_from_context(self, current_processes: Dict[int, Any]):
        if self._context.optimized:
            pids_alive = current_processes.keys()
            for pid in {*self._context.optimized.keys()}:
                if pid not in pids_alive:
                    del self._context.optimized[pid]

        if self._context.ignored_procs:
            pids_alive = current_processes.keys()
            patterns_to_remove = set()
            for pattern, ignored_ids in self._context.ignored_procs.items():
                to_remove = set()
                for id_ in ignored_ids:
                    if int(id_.split(':')[0]) not in pids_alive:
                        to_remove.add(id_)

                if to_remove:
                    self._log.debug(f"Removing dead pids from ignored context: {', '.join(f'{p}' for p in to_remove)}")
                    ignored_ids.difference_update(to_remove)

                    if not ignored_ids:
                        patterns_to_remove.add(pattern)

            if patterns_to_remove:
                for pattern in patterns_to_remove:
                    del self._context.ignored_procs[pattern]

    async def send_request(self, pid: int, command: str, match: str, profile: str):
        request = OptimizationRequest(pid=pid, command=command, created_at=time.time(),
                                      user_name=self._context.user_name, user_env=self._context.user_env,
                                      profile=profile)
        request.created_at = time.time()
        await network.send(request, self._context.opt_config, self._context.machine_id, self._log)
        self._context.optimized[pid] = match

    async def watch(self):
        while True:
            await self.check_mappings()
            await asyncio.sleep(self._context.watch_config.check_interval)



import asyncio
import getpass
import os
import time
from logging import getLogger
from typing import Tuple

from aiohttp import web

from guapow import __app_name__
from guapow.common.config import OptimizerConfig, read_optimizer_config
from guapow.common.dto import OptimizationRequest
from guapow.common.log import new_logger
from guapow.common.model_util import FileModelFiller
from guapow.service.optimizer import win_compositor, gpu
from guapow.service.optimizer.cpu import CPUFrequencyManager, get_cpu_count, CPUEnergyPolicyManager
from guapow.service.optimizer.flow import OptimizationQueue
from guapow.service.optimizer.gpu import GPUManager
from guapow.service.optimizer.handler import OptimizationHandler
from guapow.service.optimizer.mouse import MouseCursorManager
from guapow.service.optimizer.post_process.task import PostProcessTaskManager
from guapow.service.optimizer.profile import OptimizationProfile, OptimizationProfileReader, OptimizationProfileCache, \
    cache_profiles
from guapow.service.optimizer.task.manager import TasksManager, run_tasks
from guapow.service.optimizer.task.model import OptimizationContext, OptimizedProcess
from guapow.service.optimizer.watch import DeadProcessWatcherManager
from guapow.service.optimizer.web.application import create_web_app


def start():
    ti = time.time()
    app, opt_config = asyncio.get_event_loop().run_until_complete(prepare_app())
    tf = time.time()
    app.logger.info(f"Ready and attaching to port '{opt_config.port}' ({tf - ti:.4f} seconds)")
    web.run_app(port=opt_config.port, app=app, print=None)


async def prepare_app() -> Tuple[web.Application, OptimizerConfig]:
    getLogger('aiohttp.server').disabled = True

    is_service = OptimizerConfig.is_service()

    logger = new_logger(name=f'{__app_name__}-opt', service=is_service, enabled=OptimizerConfig.is_log_enabled(),
                        write_to_file=False, level=OptimizerConfig.get_log_level())
    if is_service:
        logger.debug("Initializing as a system service")

    user_id, user_name = os.getuid(), getpass.getuser()
    logger.debug(f"Initializing as user '{user_name}' (pid={os.getpid()})")

    model_filler = FileModelFiller(logger)
    opt_config = await read_optimizer_config(user_id=user_id, user_name=user_name, filler=model_filler, logger=logger)

    logger.info(f"Nice levels monitoring interval: {opt_config.renicer_interval} seconds")
    logger.info(f'Finished process checking interval: {opt_config.check_finished_interval} seconds')
    logger.info(f'Launcher mapping timeout: {opt_config.launcher_mapping_timeout} seconds')

    if opt_config.allow_root_scripts:
        logger.warning("Scripts are allowed to run at root level")

    if opt_config.profile_cache:
        logger.warning("Profile caching is enabled. Changes to files require restarting")

    compositor = None
    if opt_config.compositor:
        compositor = win_compositor.get_window_compositor_by_name(opt_config.compositor, logger)

        if compositor:
            logger.info(f'Predefined window compositor: {compositor.get_name()}')

    gpu_drivers = None

    if opt_config.gpu_vendor:
        gpu_driver = gpu.get_driver_by_vendor(opt_config.gpu_vendor)

        if gpu_driver:
            logger.info(f'Pre-defined GPU vendor: {opt_config.gpu_vendor}')
            gpu_drivers = (gpu_driver(cache=opt_config.gpu_cache, logger=logger),)
        else:
            logger.warning(f'Invalid pre-defined GPU vendor: {opt_config.gpu_vendor}')

    if opt_config.gpu_ids:
        logger.info(f"Target GPU ids to be optimized: {', '.join(str(i) for i in sorted(opt_config.gpu_ids))}")

    if not opt_config.gpu_cache:
        logger.warning("Available GPUs cache is disabled. Available GPUs will be mapped for every request")

    gpu_man = GPUManager(cache_gpus=opt_config.gpu_cache, logger=logger, drivers=gpu_drivers)

    cpu_count = get_cpu_count()
    cpufreq_man = CPUFrequencyManager(logger=logger, cpu_count=cpu_count)
    cpu_energy_man = CPUEnergyPolicyManager(logger=logger, cpu_count=cpu_count)
    context = OptimizationContext(cpufreq_man=cpufreq_man, gpu_man=gpu_man, logger=logger, cpu_count=cpu_count,
                                  compositor=compositor, allow_root_scripts=bool(opt_config.allow_root_scripts),
                                  launcher_mapping_timeout=opt_config.launcher_mapping_timeout,
                                  mouse_man=MouseCursorManager(logger), renicer_interval=opt_config.renicer_interval,
                                  cpuenergy_man=cpu_energy_man, queue=OptimizationQueue.empty(),
                                  system_service=opt_config.is_service(),
                                  gpu_ids={str(i) for i in opt_config.gpu_ids} if opt_config.gpu_ids else None)

    watcher_man = DeadProcessWatcherManager(context=context, restore_man=PostProcessTaskManager(context),
                                            check_interval=opt_config.check_finished_interval)

    tasks_man = TasksManager(context)
    await tasks_man.check_availability()

    opt_profile = OptimizationProfile.from_optimizer_config(opt_config)

    if opt_profile and opt_profile.is_valid():
        self_request = OptimizationRequest.self_request()
        self_request.prepare()

        available_init_tasks = await tasks_man.get_available_environment_tasks(OptimizedProcess(request=self_request, profile=opt_profile, created_at=time.time()))

        if available_init_tasks:
            fake_process = OptimizedProcess(request=self_request, profile=opt_profile, created_at=time.time())
            logger.debug("Waiting initial optimization tasks to complete")
            await asyncio.gather(*run_tasks(available_init_tasks, fake_process))
            logger.debug("Initial optimization tasks completed")
        else:
            logger.debug("No initial optimization tasks defined")
    else:
        logger.debug("No initial optimization tasks defined")

    profile_reader = OptimizationProfileReader(model_filler=model_filler,
                                               logger=context.logger,
                                               cache=OptimizationProfileCache(logger) if opt_config.profile_cache else None)

    if opt_config.profile_cache and opt_config.pre_cache_profiles:
        await cache_profiles(profile_reader, logger)

    handler = OptimizationHandler(context=context, tasks_man=tasks_man, watcher_man=watcher_man,
                                  profile_reader=profile_reader)
    return await create_web_app(handler=handler, queue=context.queue, config=opt_config, logger=logger), opt_config


if __name__ == '__main__':
    start()



import asyncio
import datetime
import os
import subprocess
from logging import Logger
from multiprocessing import Process
from typing import Optional, List, Dict, Set

from guapow.common import system
from guapow.common.model import ScriptSettings
from guapow.common.system import run_user_command
from guapow.common.users import is_root_user


class RunScripts:

    def __init__(self, name: str, root_allowed: bool, logger: Logger):
        self._name = name
        self._log = logger
        self.root_allowed = root_allowed

    async def _execute_user_scripts(self, settings: ScriptSettings, user_id: int, user_env: Optional[Dict[str, str]]) -> Set[int]:
        pids = set()

        for cmd in settings.scripts:
            res = system.new_user_process_response()

            self._log.info(f"Running {self._name} script '{cmd}' (user={user_id})")
            p = Process(daemon=True, target=run_user_command, args=(cmd, user_id, settings.wait_execution, settings.timeout, user_env, res))
            p.start()

            if settings.wait_execution or (settings.timeout is not None and settings.timeout > 0):
                self._log.info(f"Waiting {self._name} script '{cmd}' to finish (user={user_id})")

                while p.is_alive():  # to allow other async tasks to execute
                    await asyncio.sleep(0.0005)

                self._log.info(f"{self._name.capitalize()} script '{cmd}' finished (user={user_id})")

            p.join()

            if res['pid'] is not None:
                pids.add(res['pid'])

        return pids

    @staticmethod
    def get_environ(env: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        final_env = env if env else dict(os.environ)

        if 'DISPLAY' not in final_env:
            final_env['DISPLAY'] = ':0'

        return final_env

    async def _execute_scripts(self, settings: ScriptSettings, user_env: Optional[Dict[str, str]]) -> Set[int]:
        pids = set()
        env = self.get_environ(user_env)

        for cmd in settings.scripts:
            p = await asyncio.create_subprocess_shell(cmd=cmd, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                                                      stderr=subprocess.DEVNULL, env=env)
            pids.add(p.pid)
            self._log.info(f"Started {self._name} script: {cmd} (pid={p.pid})")

            if settings.timeout is not None:
                if not settings.has_valid_timeout():
                    self._log.warning(f"Invalid {self._name} script timeout defined: {settings.timeout}. It will not be awaited")
                    continue

                self._log.info(f"Waiting {self._name} script '{cmd}' to finish (pid={p.pid})")
                timeout = datetime.datetime.now() + datetime.timedelta(seconds=settings.timeout)

                timed_out = True
                while timeout > datetime.datetime.now():
                    if p.returncode is None:
                        await asyncio.sleep(0.001)
                    else:
                        timed_out = False
                        break

                if timed_out:
                    self._log.warning(f"{self._name.capitalize()} script '{cmd}' timed out (pid={p.pid})")

            elif settings.wait_execution:
                self._log.info(f"Waiting {self._name} script '{cmd}' to finish (pid={p.pid})")
                await p.wait()

        return pids

    async def run(self, scripts: List[ScriptSettings], user_id: Optional[int], user_env: Optional[Dict[str, str]]) -> Set[int]:
        current_user_id = os.getuid()
        root_user = is_root_user(current_user_id)

        pids = set()

        for settings in scripts:
            if root_user:
                if not settings.run_as_root and user_id is not None and not is_root_user(user_id):
                    pids.update(await self._execute_user_scripts(settings, user_id, user_env))
                elif self.root_allowed:
                    pids.update(await self._execute_scripts(settings, None))
                else:
                    self._log.warning(f"{self._name.capitalize()} scripts {settings.scripts} are not allowed to run at the root level")
            elif settings.run_as_root:
                self._log.warning(f"Cannot execute {self._name} scripts {settings.scripts} as root user")
            elif user_id is None:
                pids.update(await self._execute_scripts(settings, None))
            elif current_user_id == user_id:
                pids.update(await self._execute_scripts(settings, user_env))
            else:
                self._log.warning(f"Cannot execute {self._name} scripts {settings.scripts} as user {user_id}")

        return pids



from json import JSONEncoder


class CustomJSONEncoder(JSONEncoder):

    def default(self, o):
        if isinstance(o, set):
            return [*o]

        return super(CustomJSONEncoder, self).default(o)



from argparse import Namespace
from logging import Logger

from guapow import __app_name__
from guapow.cli.command import CLICommand
from guapow.cli.commands.service import ServiceInstaller, ServiceUninstaller

WATCHER_SERVICE_FILE = f'{__app_name__}-watch.service'


class InstallWatcher(CLICommand):

    CMD = 'install-watcher'

    def __init__(self, logger: Logger):
        super(InstallWatcher, self).__init__(logger)
        self._installer = ServiceInstaller(service_cmd=WATCHER_SERVICE_FILE.split('.service')[0],
                                           service_file=WATCHER_SERVICE_FILE, 
                                           requires_root=False,
                                           logger=logger)

    def add(self, commands: object):
        commands.add_parser(self.CMD, help="It installs and enables the watcher service (root or user level) [requires 'systemd' and 'systemctl' installed]")

    def get_command(self) -> str:
        return self.CMD

    def run(self, args: Namespace) -> bool:
        return self._installer.install()


class UninstallWatcher(CLICommand):

    CMD = 'uninstall-watcher'

    def __init__(self, logger: Logger):
        super(UninstallWatcher, self).__init__(logger)
        self._uninstaller = ServiceUninstaller(service_file=WATCHER_SERVICE_FILE, requires_root=False, logger=logger)

    def add(self, commands: object):
        commands.add_parser(self.CMD, help="It disables and uninstalls the watcher service (root or user level) [requires 'systemd' and 'systemctl' installed]")

    def get_command(self) -> str:
        return self.CMD

    def run(self, args: Namespace) -> bool:
        return self._uninstaller.uninstall()



import asyncio
from typing import Optional, List, Awaitable, Type, Dict

from guapow.service.optimizer.task.environment import EnvironmentTask, DisableWindowCompositor, \
    ChangeCPUFrequencyGovernor, ChangeGPUModeToPerformance, HideMouseCursor, StopProcessesAfterLaunch, \
    RunPostLaunchScripts, ChangeCPUEnergyPolicyLevel
from guapow.service.optimizer.task.model import OptimizationContext, Task, OptimizedProcess
from guapow.service.optimizer.task.process import ProcessTask, ReniceProcess, ChangeCPUAffinity, \
    ChangeCPUScalingPolicy, ChangeProcessIOClass


def run_tasks(tasks: List[Task], process: OptimizedProcess) -> Optional[List[Awaitable]]:
    if tasks:
        return [t.run(process) for t in tasks]


class TasksManager:

    def __init__(self, context: OptimizationContext):
        self._context = context
        self._log = context.logger
        self._proc_tasks: Optional[List[ProcessTask]] = None
        self._env_tasks: Optional[List[EnvironmentTask]] = None
        self._env_order: Dict[Type[EnvironmentTask, int]] = {
            StopProcessesAfterLaunch: 0,
            RunPostLaunchScripts: 1,
            DisableWindowCompositor: 2,
            HideMouseCursor: 3,
            ChangeCPUFrequencyGovernor: 4,
            ChangeCPUEnergyPolicyLevel: 5,
            ChangeGPUModeToPerformance: 6
        }
        self._proc_order: Dict[Type[ProcessTask, int]] = {
            ReniceProcess: 0,
            ChangeCPUAffinity: 1,
            ChangeCPUScalingPolicy: 2,
            ChangeProcessIOClass: 3
        }

    def _sort_proc(self, o: ProcessTask) -> int:
        return self._proc_order.get(o.__class__, 999)

    def _sort_env(self, o: EnvironmentTask) -> int:
        return self._env_order.get(o.__class__, 999)

    async def check_availability(self):
        self._log.debug("Checking available tasks")

        proc_opts, env_opts = [], []

        for root_cls in (EnvironmentTask, ProcessTask):
            for opt_cls in root_cls.__subclasses__():
                task = opt_cls(self._context)
                available, warning_msg = await task.is_available()

                if available:
                    if root_cls == EnvironmentTask:
                        env_opts.append(task)
                    else:
                        proc_opts.append(task)
                else:
                    self._log.warning(warning_msg)

        self._proc_tasks = proc_opts
        self._env_tasks = env_opts

        if self._proc_tasks:
            self._log.debug(f"Process tasks available ({len(self._proc_tasks)}): {', '.join([t.__class__.__name__ for t in self._proc_tasks])}")
            self._proc_tasks.sort(key=self._sort_proc)

        if self._env_tasks:
            self._log.debug(f"Environment tasks available ({len(self._env_tasks)}): {', '.join([t.__class__.__name__ for t in self._env_tasks])}")
            self._env_tasks.sort(key=self._sort_env)

    async def get_available_process_tasks(self, process: OptimizedProcess) -> Optional[List[ProcessTask]]:
        if self._proc_tasks:
            return await self._list_runnable_tasks(process, self._proc_tasks)

    async def get_available_environment_tasks(self, process: OptimizedProcess) -> Optional[List[EnvironmentTask]]:
        if self._env_tasks:
            return await self._list_runnable_tasks(process, self._env_tasks)

    async def _list_runnable_tasks(self, process: OptimizedProcess, tasks: List[Task]) -> Optional[List[Task]]:
        to_verify = [t for t in tasks if t.is_allowed_for_self_requests()] if process.request.is_self_request else tasks

        if to_verify:
            async_tasks = [self._should_run(task, process) for task in to_verify]

            if async_tasks:
                return [t for t in await asyncio.gather(*async_tasks) if t]

    async def _should_run(self, task: Task, process: OptimizedProcess) -> Optional[Task]:
        if await task.should_run(process):
            return task



import hashlib
import json
from base64 import b64encode, b64decode

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes


def encrypt(text: str, password: str) -> str:
    salt = get_random_bytes(AES.block_size)
    random_key = hashlib.scrypt(password.encode(), salt=salt, n=2**14, r=8, p=1, dklen=32)
    cipher_config = AES.new(random_key, AES.MODE_GCM)
    encrypted, tag = cipher_config.encrypt_and_digest(bytes(text, 'utf-8'))

    return b64encode(json.dumps({
        'text': b64encode(encrypted).decode('utf-8'),
        'salt': b64encode(salt).decode('utf-8'),
        'nonce': b64encode(cipher_config.nonce).decode('utf-8'),
        'tag': b64encode(tag).decode('utf-8')
    }).encode()).decode('utf-8')


def decrypt(encrypted_token: str, password: str) -> str:
    token = json.loads(b64decode(encrypted_token))
    salt = b64decode(token['salt'])
    text = b64decode(token['text'])
    nonce = b64decode(token['nonce'])
    tag = b64decode(token['tag'])

    private_key = hashlib.scrypt(password.encode(), salt=salt, n=2 ** 14, r=8, p=1, dklen=32)
    cipher = AES.new(private_key, AES.MODE_GCM, nonce=nonce)
    return cipher.decrypt_and_verify(text, tag).decode()



from guapow import __app_name__
from guapow.cli import args
from guapow.cli.commands import map_commands
from guapow.common.log import new_logger


def run():
    logger = new_logger(f'{__app_name__}-cli', service=True, enabled=True, write_to_file=False, threaded=False)
    cmds = map_commands(logger)
    current_args = args.read(cmds.values())

    selected_cmd = cmds.get(current_args.command)
    if selected_cmd:
        if not selected_cmd.run(current_args):
            exit(1)
    else:
        logger.error("No command to execute")
        exit(1)


if __name__ == '__main__':
    run()



from typing import Callable

from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import Response

from guapow.common.encryption import decrypt


@web.middleware
async def decrypt_request(request: Request, handler: Callable):
    machine_id = request.app.get('machine_id')

    if machine_id:
        try:
            encrypted_token = (await request.text()).strip()
        except:
            return Response(status=401)

        if not encrypted_token:
            return Response(status=401)

        try:
            decrypted_text = decrypt(encrypted_token, machine_id)
        except:
            return Response(status=401)

        if not decrypted_text:
            return Response(status=401)

        request['decrypted_text'] = decrypted_text

    return await handler(request)


@web.middleware
async def hide_server_headers(request: Request, handler: Callable):
    response = await handler(request)
    response.headers['Server'] = ''
    return response



import asyncio
import os
import time
from typing import Optional, Awaitable, List

from guapow.common.dto import OptimizationRequest
from guapow.common.profile import get_possible_profile_paths_by_priority, \
    get_default_profile_name
from guapow.service.optimizer.launcher import LauncherMapperManager
from guapow.service.optimizer.profile import OptimizationProfile, OptimizationProfileReader
from guapow.service.optimizer.task.manager import TasksManager, run_tasks
from guapow.service.optimizer.task.model import OptimizationContext, OptimizedProcess
from guapow.service.optimizer.watch import DeadProcessWatcherManager


class OptimizationHandler:

    def __init__(self, context: OptimizationContext, tasks_man: TasksManager, watcher_man: DeadProcessWatcherManager,
                 profile_reader: OptimizationProfileReader):
        super(OptimizationHandler, self)
        self.context = context
        self._queue = context.queue
        self._log = context.logger
        self._tasks_man = tasks_man
        self._watcher_man = watcher_man
        self._profile_reader = profile_reader
        self._launcher_mapper = LauncherMapperManager(check_time=context.launcher_mapping_timeout, logger=context.logger)

    async def _read_valid_profile(self, name: str, add_settings: Optional[str], user_id: Optional[int], user_name: Optional[str], request: OptimizationRequest) -> Optional[OptimizationProfile]:
        for file_path in get_possible_profile_paths_by_priority(name=name, user_id=user_id, user_name=user_name):
            if file_path:
                try:
                    return await self._profile_reader.read_valid(profile_path=file_path, add_settings=add_settings, handle_not_found=False)
                except FileNotFoundError:
                    self._log.debug(f"Profile file '{file_path}' not found (request={request.pid})")

    async def _load_valid_profile(self, request: OptimizationRequest) -> Optional[OptimizationProfile]:
        profile = None

        if request.profile:
            profile = await self._read_valid_profile(request.profile, request.profile_config, request.user_id, request.user_name, request)

            if profile:
                self._log.info(f"Valid profile '{profile.name}' ({profile.path}) found (request={request.pid})")

        if not profile:
            profile = await self._read_valid_profile(get_default_profile_name(), request.profile_config, request.user_id, request.user_name, request)

            if profile:
                pre_msg = "No existing/valid profile '{}'. ".format(request.profile) if request.profile else "Request has no profile defined. "
                self._log.warning(f"{pre_msg}Profile '{profile.path}' will be used instead (request={request.pid})")

        return profile

    def map_valid_config(self, config: str) -> Optional[OptimizationProfile]:
        if config:
            profile = self._profile_reader.map(profile_str=config)

            if profile and profile.is_valid():
                return profile

        self._log.warning("No optimization settings defined in configuration: {}".format(config.replace('\n', ' ')))

    async def _start_environment_tasks(self, process: OptimizedProcess) -> Optional[List[Awaitable]]:
        env_tasks = await self._tasks_man.get_available_environment_tasks(process)

        if env_tasks:
            return run_tasks(env_tasks, process)

    async def _start_process_tasks(self, process: OptimizedProcess) -> Optional[List[Awaitable]]:
        if process.profile.process:
            proc_tasks = await self._tasks_man.get_available_process_tasks(process)

            if proc_tasks:
                mapped_pid = await self._launcher_mapper.map_pid(process.request, process.profile)

                if mapped_pid is not None:
                    process.pid = mapped_pid
                    await self._queue.add_pid(mapped_pid)

                return run_tasks(proc_tasks, process)

    async def handle(self, request: OptimizationRequest):
        request.prepare()

        if not os.path.exists(f'/proc/{request.pid}'):
            self._log.warning(f'Process {request.pid} does not exist. No optimization will be applied.')
            await self.context.queue.remove_pids(request.pid)
        else:
            if request.has_full_configuration():
                profile = self.map_valid_config(request.config)
            else:
                profile = await self._load_valid_profile(request)

            if not profile:
                self._log.warning(f"No optimizations available for process '{request.pid}'")

            process = OptimizedProcess(request=request, created_at=time.time(), profile=profile)

            proc_tasks_running = None

            if profile:
                env_tasks_running = await self._start_environment_tasks(process)
                proc_tasks_running = await self._start_process_tasks(process)

                if env_tasks_running:
                    await asyncio.gather(*env_tasks_running)

            should_be_watched = process.should_be_watched()

            if should_be_watched:
                await self._watcher_man.watch(process)

                if process.pid != process.source_pid:
                    await self._queue.remove_pids(process.source_pid)

            if proc_tasks_running:
                await asyncio.gather(*proc_tasks_running)

            if not should_be_watched:
                related_pids = process.get_pids()

                if related_pids:
                    await self._queue.remove_pids(*related_pids)

            exec_time = time.time() - request.created_at
            self._log.debug(f"Optimization request for '{request.pid}' took {exec_time:.4f} seconds"
                            f"{f' (target_pid={process.pid})' if request.pid != process.pid else ''}")



from abc import ABC, abstractmethod
from argparse import Namespace
from logging import Logger


class CLICommand(ABC):

    @abstractmethod
    def __init__(self, logger: Logger):
        self._log = logger

    @abstractmethod
    def add(self, commands: object):
        pass

    @abstractmethod
    def get_command(self) -> str:
        pass

    @abstractmethod
    def run(self, args: Namespace) -> bool:
        pass



import asyncio
import os
from logging import Logger
from typing import Optional

from guapow.common import system


class Renicer:

    def __init__(self, logger: Logger, watch_interval: float):
        self._log = logger
        self._pid_nice = {}
        self._watching = False
        self._watch_interval = watch_interval

    def get_priority(self, pid: int) -> Optional[int]:
        try:
            return os.getpriority(os.PRIO_PROCESS, pid)
        except:
            pass

    def set_priority(self, pid: int, level: int, request_pid: int) -> bool:
        try:
            os.setpriority(os.PRIO_PROCESS, pid, level)
            self._log.info(f"Process {pid} nice level changed to '{level}' (request={request_pid})")
            return True
        except:
            self._log.error(f"Could not change process {pid} nice level to {level} (request={request_pid})")
            return False

    def add(self, pid: int, nice_level: int, request_pid: int) -> bool:
        if pid not in self._pid_nice:
            self._pid_nice[pid] = nice_level, request_pid
            self._log.info(f"Process {pid} nice level will be monitored (request={request_pid})")
            return True

        self._log.debug(f"Process {pid} nice level is already being monitored (request={request_pid})")
        return False

    def is_watching(self) -> bool:
        return self._watching

    async def _watch(self):
        while True:
            if not self._pid_nice:
                break

            pids = system.read_current_pids()
            dead_pids = set()

            for pid, nice_request in self._pid_nice.items():
                level, request_pid = nice_request[0], nice_request[1]
                if pid not in pids:
                    dead_pids.add(pid)

                current_nice = self.get_priority(pid)

                if current_nice != nice_request[0]:
                    self._log.debug(
                        f"Process {pid} current nice level ({current_nice}) differs from expected ({level}) "
                        f"(request={request_pid})")

                    self.set_priority(pid, level, request_pid)

            if dead_pids:
                for pid in dead_pids:
                    if pid in self._pid_nice:
                        del self._pid_nice[pid]

                self._log.debug(f"Stop monitoring the nice level of processes: {', '.join((str(p) for p in dead_pids))}")

            if not self._pid_nice:
                break

            if self._watch_interval and self._watch_interval > 0:
                await asyncio.sleep(self._watch_interval)

        self._watching = False
        self._log.debug(f"Stopped monitoring nice levels")

    def watch(self) -> bool:
        if not self._watching and self._pid_nice:
            self._watching = True
            asyncio.get_event_loop().create_task(self._watch())
            return True

        return False



import os
from logging import Logger
from typing import Optional, Dict, Tuple

import aiofiles

from guapow import __app_name__
from guapow.common.profile import get_default_profile_name
from guapow.common.users import is_root_user

FILE_NAME = 'watch.map'


def get_root_file_path() -> str:
    return f'/etc/{__app_name__}/{FILE_NAME}'


def get_user_file_path(user_name: str) -> str:
    return f'/home/{user_name}/.config/{__app_name__}/{FILE_NAME}'


def get_file_path(user_id: int, user_name: str) -> Optional[str]:
    return get_root_file_path() if is_root_user(user_id) else get_user_file_path(user_name)


def get_existing_file_path(user_id: int, user_name: str, logger: Logger) -> Optional[str]:
    file_path = get_file_path(user_id, user_name)
    if os.path.isfile(file_path):
        logger.info(f"Mapping file '{file_path}' found")
        return file_path
    else:
        logger.warning(f"Mapping file '{file_path}' not found")
        return get_existing_file_path(0, 'root', logger) if not is_root_user(user_id) else None


def get_default_file_path(user_id: int, user_name: str, logger: Logger) -> str:
    file_path = get_existing_file_path(user_id, user_name, logger)

    if file_path:
        return file_path

    default_path = get_file_path(user_id, user_name)
    logger.info(f'Considering the default mapping file path for current user: {default_path}')
    return default_path


async def read(file_path: str, logger: Logger, last_file_found_log: Optional[bool]) -> Tuple[bool, Optional[Dict[str, str]]]:
    try:
        async with aiofiles.open(file_path) as f:
            mapping_file_str = (await f.read()).strip()

        if not last_file_found_log:
            logger.error(f"Mapping file '{file_path}' found")

    except FileNotFoundError:
        if last_file_found_log is None or last_file_found_log is True:
            logger.error(f"Mapping file '{file_path}' not found")

        return False, None

    mappings = map_string(mapping_file_str)
    return True, mappings if mappings else None


def map_string(string: str) -> Optional[Dict[str, str]]:
    if string:
        mappings = {}
        for line in string.split('\n'):
            if line:
                line_strip = line.strip()

                if line_strip:
                    line_no_comment = [w.strip() for w in line_strip.split('#', 1)]

                    if not line_no_comment or not line_no_comment[0]:
                        continue

                    try:
                        div_idx = line_no_comment[0].rindex('=')
                        pattern = line_no_comment[0][0:div_idx].strip()
                        profile = line_no_comment[0][div_idx + 1:].strip()
                    except ValueError:
                        pattern = line_no_comment[0].strip()
                        profile = None

                    if pattern and not profile:
                        mappings[pattern] = get_default_profile_name()
                    else:
                        mappings[pattern] = profile

        if mappings:
            return mappings
